# -*- coding: UTF-8 -*-
"""
Capy Puzzle
"""

from dataclasses import dataclass
from typing import Optional

from enforce_typing import enforce_types  # type: ignore

from .base import BaseCaptcha, BaseCaptchaSolution


@enforce_types
@dataclass
class CapyPuzzle(BaseCaptcha):
    """ Capy Puzzle CAPTCHA """

    site_key: str
    page_url: str
    api_server: Optional[str] = None


@enforce_types
@dataclass
class CapyPuzzleSolution(BaseCaptchaSolution):
    """ Capy Puzzle CAPTCHA solution """

    captchakey: str
    challengekey: str
    answer: str



# -*- coding: UTF-8 -*-
"""
CAPTCHA common stuff
"""

import enum


class CaptchaAlphabet(enum.Enum):
    """ Alphabet used in the CAPTCHA """

    LATIN = 'latin'
    CYRILLIC = 'cyrillic'


class CaptchaCharType(enum.Enum):
    """ Character types used in CAPTCHA """

    NUMERIC = 1
    ALPHA = 2
    ALPHA_OR_NUMERIC = 3
    ALPHANUMERIC = 4


class WorkerLanguage(enum.Enum):
    """ Worker's language to solve the CAPTCHA """

    ENGLISH = 'en'
    RUSSIAN = 'ru'
    SPANISH = 'es'
    PORTUGUESE = 'pt'
    UKRAINIAN = 'uk'
    VIETNAMESE = 'vi'
    FRENCH = 'fr'
    INDONESIAN = 'id'
    ARAB = 'ar'
    JAPANESE = 'ja'
    TURKISH = 'tr'
    GERMAN = 'de'
    CHINESE = 'zh'
    # PHILIPPINE = 'fil'
    POLISH = 'pl'
    THAI = 'th'
    ITALIAN = 'it'
    DUTCH = 'nl'
    SLOVAK = 'sk'
    BULGARIAN = 'bg'
    ROMANIAN = 'ro'
    HUNGARIAN = 'hu'
    KOREAN = 'ko'
    CZECH = 'cs'
    AZERBAIJANI = 'az'
    PERSIAN = 'fa'
    BENGALI = 'bn'
    GREEK = 'el'
    LITHUANIAN = 'lt'
    LATVIAN = 'lv'
    SWEDISH = 'sv'
    SERBIAN = 'sr'
    CROATIAN = 'hr'
    HEBREW = 'he'
    HINDI = 'hi'
    NORWEGIAN = 'nb'
    SLOVENIAN = 'sl'
    DANISH = 'da'
    UZBEK = 'uz'
    FINNISH = 'fi'
    CATALAN = 'ca'
    GEORGIAN = 'ka'
    MALAY = 'ms'
    TELUGU = 'te'
    ESTONIAN = 'et'
    MALAYALAM = 'ml'
    BELORUSSIAN = 'be'
    KAZAKH = 'kk'
    MARATHI = 'mr'
    NEPALI = 'ne'
    BURMESE = 'my'
    BOSNIAN = 'bs'
    ARMENIAN = 'hy'
    MACEDONIAN = 'mk'
    PUNJABI = 'pa'



# -*- coding: UTF-8 -*-

import base64
import os.path
import pathlib

import pytest
from unicaps.captcha import ImageCaptcha
from unicaps.exceptions import BadInputDataError

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
IMAGE_FILE = os.path.join(DATA_DIR, 'image.jpg')


@pytest.fixture
def image_path():
    return pathlib.Path(IMAGE_FILE)


@pytest.fixture
def image_bytes():
    return open(IMAGE_FILE, 'rb').read()


@pytest.fixture
def image_base64(image_bytes):
    return base64.b64encode(image_bytes)


def test_image_from_path(image_path, image_bytes):
    """ Pathlib input test"""
    captcha = ImageCaptcha(image=image_path)
    assert image_bytes == captcha.get_image_bytes()


def test_image_from_binary(image_bytes):
    """ Binary input test"""
    captcha = ImageCaptcha(image=image_bytes)
    assert image_bytes == captcha.get_image_bytes()


def test_not_an_image():
    """ Not an image input test """
    with pytest.raises(TypeError):
        ImageCaptcha(image='wrong_image')


def test_bad_image():
    """ Bad image input test """
    with pytest.raises(BadInputDataError):
        ImageCaptcha(image=b'bad_image_data')



# -*- coding: UTF-8 -*-
"""
Google reCAPTCHA v2
"""

from dataclasses import dataclass
from typing import Optional

from enforce_typing import enforce_types  # type: ignore

from .base import BaseCaptcha, BaseCaptchaSolution


@enforce_types
@dataclass
class RecaptchaV2(BaseCaptcha):
    """ Google reCAPTCHA v2 """

    site_key: str
    page_url: str
    is_invisible: bool = False
    is_enterprise: bool = False
    data_s: Optional[str] = None
    api_domain: Optional[str] = None


@enforce_types
@dataclass
class RecaptchaV2Solution(BaseCaptchaSolution):
    """ Google reCAPTCHA v2 solution """

    token: str



# -*- coding: UTF-8 -*-
"""
KeyCaptcha
"""

from dataclasses import dataclass

from enforce_typing import enforce_types  # type: ignore

from .base import BaseCaptcha, BaseCaptchaSolution


@enforce_types
@dataclass
class KeyCaptcha(BaseCaptcha):
    """ KeyCaptcha """

    page_url: str
    user_id: str
    session_id: str
    ws_sign: str
    ws_sign2: str


@enforce_types
@dataclass
class KeyCaptchaSolution(BaseCaptchaSolution):
    """ KeyCaptcha solution """

    token: str



# -*- coding: UTF-8 -*-
"""
hCaptcha
"""

from dataclasses import dataclass
from typing import Optional

from enforce_typing import enforce_types  # type: ignore

from .base import BaseCaptcha, BaseCaptchaSolution


@enforce_types
@dataclass
class HCaptcha(BaseCaptcha):
    """ hCaptcha """

    site_key: str
    page_url: str
    is_invisible: bool = False
    api_domain: Optional[str] = None


@enforce_types
@dataclass
class HCaptchaSolution(BaseCaptchaSolution):
    """ hCaptcha solution """

    token: str



# -*- coding: UTF-8 -*-
"""
Base service stuff
"""

import asyncio
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from inspect import getmodule
from timeit import default_timer as timer
from typing import Dict, Optional, Tuple

from .._transport.http_transport import StandardHTTPTransport  # type: ignore
from .._captcha import CaptchaType
from .._captcha.base import BaseCaptcha, BaseCaptchaSolution
from .._misc.proxy import ProxyServer
from ..exceptions import UnicapsException, SolutionWaitTimeout, SolutionNotReadyYet


class BaseService(ABC):
    """ Base class for all services """

    def __init__(self, api_key: str):
        self.api_key = api_key
        self._transport = self._init_transport()
        self._module = getmodule(self)
        self._settings = {captcha_type: Settings() for captcha_type in self.supported_captchas}
        self._post_init()

    @abstractmethod
    def _init_transport(self):
        pass

    def _post_init(self):
        pass

    def _make_request(self, request_class, *args):
        request_class = request_class + "Request"
        if not hasattr(self._module, request_class):
            raise UnicapsException(f"{request_class} is not supported by the current service!")

        request = getattr(self._module, request_class)(self)
        return self._transport.make_request(request, *args)

    async def _make_request_async(self, request_class, *args):
        request_class = request_class + "Request"
        if not hasattr(self._module, request_class):
            raise UnicapsException(f"{request_class} is not supported by the current service!")

        request = getattr(self._module, request_class)(self)
        return await self._transport.make_request_async(request, *args)

    @property
    def supported_captchas(self) -> Tuple[CaptchaType, ...]:
        """ List of supported captchas """

        captchas = []
        for captcha_type in CaptchaType:
            if hasattr(self._module, captcha_type.value + "TaskRequest"):
                captchas.append(captcha_type)
        return tuple(captchas)

    @property
    def settings(self) -> Dict[CaptchaType, 'Settings']:
        """ Service settings """

        return self._settings

    def solve_captcha(self, captcha: BaseCaptcha, proxy: Optional[ProxyServer] = None,
                      user_agent: Optional[str] = None,
                      cookies: Optional[Dict[str, str]] = None) -> 'SolvedCaptcha':
        """ Solves captcha and returns SolvedCaptcha object """

        start_time = datetime.now()
        task = self.create_task(captcha, proxy, user_agent, cookies)
        solution, cost, extra = self.wait_for_solution(task)
        end_time = datetime.now()

        return SolvedCaptcha(task, solution, start_time, end_time,
                             cost=cost, extra=extra)

    async def solve_captcha_async(self, captcha: BaseCaptcha, proxy: Optional[ProxyServer] = None,
                                  user_agent: Optional[str] = None,
                                  cookies: Optional[Dict[str, str]] = None) -> 'AsyncSolvedCaptcha':
        """ Solves captcha and returns SolvedCaptcha object (async) """

        start_time = datetime.now()
        task = await self.create_task_async(captcha, proxy, user_agent, cookies)
        solution, cost, extra = await self.wait_for_solution_async(task)
        end_time = datetime.now()

        return AsyncSolvedCaptcha(task, solution, start_time, end_time,
                                  cost=cost, extra=extra)

    def create_task(self, captcha: BaseCaptcha, proxy: Optional[ProxyServer] = None,
                    user_agent: Optional[str] = None,
                    cookies: Optional[Dict[str, str]] = None) -> 'CaptchaTask':
        """ Creates task for solving a CAPTCHA """

        captcha_type = captcha.get_type()

        if captcha_type not in self.supported_captchas:
            raise UnicapsException(f"{captcha_type} is not supported by the current service!")

        result = self._make_request(
            f"{captcha_type.value}Task", captcha, proxy, user_agent, cookies
        )
        task_id = str(result["task_id"])

        return CaptchaTask(self, captcha, task_id, result.get("extra"))

    async def create_task_async(self, captcha: BaseCaptcha, proxy: Optional[ProxyServer] = None,
                                user_agent: Optional[str] = None,
                                cookies: Optional[Dict[str, str]] = None) -> 'AsyncCaptchaTask':
        """ Creates CAPTCHA solving task (async) """

        captcha_type = captcha.get_type()

        if captcha_type not in self.supported_captchas:
            raise UnicapsException(f"{captcha_type} is not supported by the current service!")

        result = await self._make_request_async(
            f"{captcha_type.value}Task", captcha, proxy, user_agent, cookies
        )
        task_id = str(result["task_id"])

        return AsyncCaptchaTask(self, captcha, task_id, result.get("extra"))

    def get_task_result(self, task: 'CaptchaTask') -> Tuple[BaseCaptchaSolution,
                                                            Optional[float], Dict]:
        """ Returns CAPTCHA solution """

        result = self._make_request(f"{task.captcha.get_type().value}Solution", task)

        return (
            result['solution'],  # type: ignore
            float(result['cost']) if result.get('cost') else None,
            result.get("extra") or {}
        )

    async def get_task_result_async(self, task: 'CaptchaTask') -> Tuple[BaseCaptchaSolution,
                                                                        Optional[float], Dict]:
        """ Returns CAPTCHA solution """

        result = await self._make_request_async(f"{task.captcha.get_type().value}Solution", task)

        return (
            result['solution'],  # type: ignore
            float(result['cost']) if result.get('cost') else None,
            result.get("extra") or {}
        )

    def wait_for_solution(self, task) -> Tuple[BaseCaptchaSolution, Optional[float], Dict]:
        """ Wait for CAPTCHA solution """

        settings = self._settings[task.captcha.get_type()]

        start_time = timer()
        time.sleep(settings.polling_delay)
        while True:
            if timer() - start_time > settings.solution_timeout:
                raise SolutionWaitTimeout(
                    f"Couldn't receive a solution in {settings.solution_timeout} seconds!"
                )

            try:
                return task.get_result()
            except SolutionNotReadyYet:
                time.sleep(settings.polling_interval)

    async def wait_for_solution_async(self, task) -> Tuple[BaseCaptchaSolution,
                                                           Optional[float], Dict]:
        """ Wait for CAPTCHA solution """

        settings = self._settings[task.captcha.get_type()]

        start_time = timer()
        await asyncio.sleep(settings.polling_delay)
        while True:
            if timer() - start_time > settings.solution_timeout:
                raise SolutionWaitTimeout(
                    f"Couldn't receive a solution in {settings.solution_timeout} seconds!"
                )

            try:
                return await task.get_result()
            except SolutionNotReadyYet:
                await asyncio.sleep(settings.polling_interval)

    def get_balance(self):
        """ Get account balance """

        balance = self._make_request("GetBalance").get('balance')
        if balance is not None:
            balance = float(balance)
        return balance

    async def get_balance_async(self):
        """ Get account balance """

        response = await self._make_request_async("GetBalance")
        balance = response.get('balance')
        if balance is not None:
            balance = float(balance)
        return balance

    def get_status(self) -> bool:
        """ Get service status """

        return bool(self._make_request("GetStatus"))

    async def get_status_async(self) -> bool:
        """ Get service status """

        return bool(await self._make_request_async("GetStatus"))

    def report_good(self, solved_captcha: 'SolvedCaptcha', raise_exc: bool = False) -> bool:
        """ Report good CAPTCHA """

        result = False
        try:
            result = self._make_request("ReportGood", solved_captcha)
        except UnicapsException:
            if raise_exc:
                raise
        return bool(result)

    async def report_good_async(self, solved_captcha: 'SolvedCaptcha',
                                raise_exc: bool = False) -> bool:
        """ Report good CAPTCHA """

        result = False
        try:
            result = await self._make_request_async("ReportGood", solved_captcha)
        except UnicapsException:
            if raise_exc:
                raise
        return bool(result)

    def report_bad(self, solved_captcha: 'SolvedCaptcha', raise_exc: bool = False) -> bool:
        """ Report bad CAPTCHA """

        result = False
        try:
            result = self._make_request("ReportBad", solved_captcha)
        except UnicapsException:
            if raise_exc:
                raise
        return bool(result)

    async def report_bad_async(self, solved_captcha: 'SolvedCaptcha',
                               raise_exc: bool = False) -> bool:
        """ Report bad CAPTCHA """

        result = False
        try:
            result = await self._make_request_async("ReportBad", solved_captcha)
        except UnicapsException:
            if raise_exc:
                raise
        return bool(result)

    @abstractmethod
    def close(self):
        """ Close connections """

    @abstractmethod
    async def close_async(self):
        """ Close connections (async) """


class HTTPService(BaseService):
    """ Standard HTTP Service """

    def _init_transport(self):
        return StandardHTTPTransport()

    def close(self):
        """ Close connections """
        self._transport.close()

    async def close_async(self):
        """ Close connections (async) """
        await self._transport.close_async()


@dataclass
class Settings:
    """ Service settings """

    polling_delay: int = 5  # seconds before starting to check for sollution
    polling_interval: int = 2  # seconds between checks
    solution_timeout: int = 300  # seconds is solution timeout


class CaptchaTask:
    """ Task for CAPTCHA solving """

    def __init__(self, service, captcha: BaseCaptcha, task_id: str, extra: Dict = None):
        self._service = service
        self._captcha = captcha
        self._task_id = task_id
        self._extra = extra or {}
        self._result = None

    @property
    def task_id(self) -> str:
        """ Task ID """
        return self._task_id

    @property
    def captcha(self) -> BaseCaptcha:
        """ Source CAPTCHA """
        return self._captcha

    @property
    def extra(self) -> Dict:
        """ Task extra data """
        return self._extra

    def get_result(self) -> Optional[BaseCaptchaSolution]:
        """ Gets solution """
        if self._result is None:
            self._result = self._service.get_task_result(self)
        return self._result

    def is_done(self) -> bool:
        """ Checks if solution is ready """
        return bool(self._result)

    def wait(self) -> BaseCaptchaSolution:
        """ Waits for solution """
        return self._service.wait_for_solution(self)


class AsyncCaptchaTask(CaptchaTask):
    """ Task for CAPTCHA solving """

    async def get_result(self) -> Optional[BaseCaptchaSolution]:  # type: ignore
        """ Gets solution """
        if self._result is None:
            self._result = await self._service.get_task_result_async(self)
        return self._result

    async def wait(self) -> BaseCaptchaSolution:  # type: ignore
        """ Waits for solution """
        return await self._service.wait_for_solution_async(self)


class SolvedCaptcha:
    """ Solved CAPTCHA object """

    def __init__(self, task: CaptchaTask, solution: BaseCaptchaSolution, start_time: datetime,
                 end_time: datetime, cost: Optional[float] = None, cookies: Optional[dict] = None,
                 extra: dict = None):
        if not task.is_done():
            raise UnicapsException("CAPTCHA is not solved yet!")

        self._task = task
        self._solution = solution
        self._start_time = start_time
        self._end_time = end_time
        self._cost = cost
        self._cookies = cookies or {}
        self._extra = extra or {}

    @property
    def captcha_id(self) -> str:
        """ CAPTCHA ID (usually it's the same as task ID) """
        return self._task.task_id

    @property
    def task(self) -> CaptchaTask:
        """ Task for solving """
        return self._task

    @property
    def solution(self) -> BaseCaptchaSolution:
        """ CAPTCHA solution """
        return self._solution

    @property
    def start_time(self) -> datetime:
        """ Start solving at """
        return self._start_time

    @property
    def end_time(self) -> datetime:
        """ End solving at """
        return self._end_time

    @property
    def cost(self) -> Optional[float]:
        """ The cost of solved CAPTCHA """
        return self._cost

    @property
    def cookies(self) -> dict:
        """ Cookies """
        return self._cookies

    @property
    def extra(self) -> dict:
        """ Extra data from the service """
        return self._extra

    def report_good(self, raise_exc: bool = False) -> bool:
        """ Report good CAPTCHA """
        # pylint: disable=protected-access
        return self._task._service.report_good(self, raise_exc=raise_exc)

    def report_bad(self, raise_exc: bool = False) -> bool:
        """ Report bad CAPTCHA """
        # pylint: disable=protected-access
        return self._task._service.report_bad(self, raise_exc=raise_exc)


class AsyncSolvedCaptcha(SolvedCaptcha):
    """ Solved CAPTCHA object (async) """

    async def report_good(self, raise_exc: bool = False) -> bool:  # type: ignore
        """ Report good CAPTCHA """
        # pylint: disable=protected-access
        return await self._task._service.report_good_async(self, raise_exc=raise_exc)

    async def report_bad(self, raise_exc: bool = False) -> bool:  # type: ignore
        """ Report bad CAPTCHA """
        # pylint: disable=protected-access
        return await self._task._service.report_bad_async(self, raise_exc=raise_exc)



"""
captcha.guru service
"""

# pylint: disable=unused-import
from .twocaptcha import (
    Service as Service2Captcha, GetBalanceRequest, GetStatusRequest,
    ReportGoodRequest, ReportBadRequest,
    ImageCaptchaTaskRequest, ImageCaptchaSolutionRequest,
    RecaptchaV2TaskRequest, RecaptchaV2SolutionRequest,
    RecaptchaV3TaskRequest, RecaptchaV3SolutionRequest,
    HCaptchaTaskRequest, HCaptchaSolutionRequest,
    GeeTestTaskRequest, GeeTestSolutionRequest
)

__all__ = [
    'Service', 'GetBalanceRequest', 'GetStatusRequest',
    'ReportGoodRequest', 'ReportBadRequest',
    'ImageCaptchaTaskRequest', 'ImageCaptchaSolutionRequest',
    'RecaptchaV2TaskRequest', 'RecaptchaV2SolutionRequest',
    'RecaptchaV3TaskRequest', 'RecaptchaV3SolutionRequest',
    'HCaptchaTaskRequest', 'HCaptchaSolutionRequest',
    'GeeTestTaskRequest', 'GeeTestSolutionRequest'
]


class Service(Service2Captcha):
    """ Main service class for captcha.guru """

    BASE_URL = 'http://api.captcha.guru'


def _decorator(cls):
    """ Decorator for *TaskRequest class """

    # pylint: disable=missing-function-docstring
    class Wrapper:
        """ A wrapper for *TaskRequest class """
        def __init__(self, *args, **kwargs):
            self.decorated_obj = cls(*args, **kwargs)

        def prepare(self, *args, **kwargs):
            result = self.decorated_obj.prepare(*args, **kwargs)
            if 'data' in result:
                result['params'] = result.pop('data')
                result['method'] = 'GET'

                if 'soft_id' in result['params']:
                    del result['params']['soft_id']
                    result['params']['softguru'] = '127872'

            return result

        def parse_response(self, *args, **kwargs):
            return self.decorated_obj.parse_response(*args, **kwargs)

        def process_response(self, *args, **kwargs):
            return self.decorated_obj.process_response(*args, **kwargs)

    return Wrapper


ImageCaptchaTaskRequest = _decorator(ImageCaptchaTaskRequest)  # type: ignore
RecaptchaV2TaskRequest = _decorator(RecaptchaV2TaskRequest)  # type: ignore
RecaptchaV3TaskRequest = _decorator(RecaptchaV3TaskRequest)  # type: ignore
HCaptchaTaskRequest = _decorator(HCaptchaTaskRequest)  # type: ignore
GeeTestTaskRequest = _decorator(GeeTestTaskRequest)  # type: ignore



# -*- coding: UTF-8 -*-
"""
GeeTest v4 CAPTCHA
"""

from dataclasses import dataclass
from typing import Optional

from enforce_typing import enforce_types  # type: ignore

from .base import BaseCaptcha, BaseCaptchaSolution


@enforce_types
@dataclass
class GeeTestV4(BaseCaptcha):
    """ GeeTest v4 """

    page_url: str
    captcha_id: str


@enforce_types
@dataclass
class GeeTestV4Solution(BaseCaptchaSolution):
    """ GeeTest v4 solution """

    captcha_id: str
    lot_number: str
    pass_token: str
    gen_time: str
    captcha_output: str



# -*- coding: UTF-8 -*-
"""
Proxy tests
"""

from unicaps.proxy import ProxyServer, ProxyServerType


def test_proxy_from_string():
    proxy = ProxyServer('http://login:password@address:8080')

    assert proxy.address == 'address'
    assert proxy.proxy_type == ProxyServerType.HTTP
    assert proxy.port == 8080
    assert proxy.login == 'login'
    assert proxy.password == 'password'


def test_proxy_wo_auth_from_string():
    proxy = ProxyServer('https://address:8080')

    assert proxy.address == 'address'
    assert proxy.proxy_type == ProxyServerType.HTTPS
    assert proxy.port == 8080
    assert proxy.login is None
    assert proxy.password is None


def test_proxy_address_only():
    proxy = ProxyServer('address')

    assert proxy.address == 'address'
    assert proxy.proxy_type == ProxyServerType.HTTP
    assert proxy.port == 80
    assert proxy.login is None
    assert proxy.password is None


def test_proxy_all():
    proxy = ProxyServer('address', ProxyServerType.SOCKS4, 80, 'login', 'password')

    assert proxy.address == 'address'
    assert proxy.proxy_type == ProxyServerType.SOCKS4
    assert proxy.port == 80
    assert proxy.login == 'login'
    assert proxy.password == 'password'


def test_proxy_to_string():
    proxy = ProxyServer('address', ProxyServerType.SOCKS5, 80, 'login', 'password')

    assert str(proxy) == 'socks5://login:password@address:80'



# -*- coding: UTF-8 -*-
"""
Google reCAPTCHA v3
"""

from dataclasses import dataclass
from typing import Optional

from enforce_typing import enforce_types  # type: ignore

from .base import BaseCaptcha, BaseCaptchaSolution


@enforce_types
@dataclass
class RecaptchaV3(BaseCaptcha):
    """ Google reCAPTCHA v3 """

    site_key: str
    page_url: str
    is_enterprise: bool = False
    action: Optional[str] = None
    min_score: Optional[float] = None
    api_domain: Optional[str] = None


@enforce_types
@dataclass
class RecaptchaV3Solution(BaseCaptchaSolution):
    """ Google reCAPTCHA v3 solution """

    token: str



# -*- coding: UTF-8 -*-
"""
anti-captcha.com service
"""

import json

from .base import HTTPService
from .._transport.http_transport import HTTPRequestJSON  # type: ignore
from .. import exceptions
from .._captcha import CaptchaType
from ..common import WorkerLanguage

__all__ = [
    'Service', 'GetBalanceRequest', 'GetStatusRequest',
    'ReportGoodRequest', 'ReportBadRequest',
    'ImageCaptchaTaskRequest', 'ImageCaptchaSolutionRequest',
    'RecaptchaV2TaskRequest', 'RecaptchaV2SolutionRequest',
    'RecaptchaV3TaskRequest', 'RecaptchaV3SolutionRequest',
    'FunCaptchaTaskRequest', 'FunCaptchaSolutionRequest',
    'GeeTestTaskRequest', 'GeeTestSolutionRequest',
    'GeeTestV4TaskRequest', 'GeeTestV4SolutionRequest',
    'HCaptchaTaskRequest', 'HCaptchaSolutionRequest',
]


class Service(HTTPService):
    """ Main service class for anti-captcha """

    BASE_URL = 'https://api.anti-captcha.com'

    def _post_init(self):
        """ Init settings """

        for captcha_type in self.settings:
            self.settings[captcha_type].polling_interval = 2

            if captcha_type in (CaptchaType.IMAGE,):
                self.settings[captcha_type].polling_delay = 5
                self.settings[captcha_type].solution_timeout = 90
            else:
                self.settings[captcha_type].polling_delay = 10
                self.settings[captcha_type].solution_timeout = 300


class Request(HTTPRequestJSON):
    """ Common Request class for anti-captcha """

    def prepare(self, **kwargs) -> dict:
        """ Prepares request """

        request = super().prepare(**kwargs)
        request.update(
            dict(
                method="POST",
                json=dict(clientKey=self._service.api_key)
            )
        )
        return request

    def parse_response(self, response) -> dict:
        """ Parses response and checks for errors """

        response_data = super().parse_response(response)

        error_id = response_data.pop("errorId")
        if error_id == 0:
            return response_data

        # ############# #
        # handle errors #
        # ############# #
        error_code = response_data.get("errorCode", f'ERROR {error_id}')
        error_text = response_data.get("errorDescription", "")
        error_msg = f"{error_code}: {error_text}"

        # pylint: disable=no-else-raise
        if error_code in ('ERROR_WRONG_USER_KEY', 'ERROR_KEY_DOES_NOT_EXIST',
                          'ERROR_IP_NOT_ALLOWED', 'ERROR_IP_BLOCKED'):
            raise exceptions.AccessDeniedError(error_msg)
        elif error_code in ('ERROR_ZERO_BALANCE',):
            raise exceptions.LowBalanceError(error_msg)
        elif error_code in ('ERROR_NO_SLOT_AVAILABLE',):
            raise exceptions.ServiceTooBusy(error_msg)
        elif error_code in ('ERROR_NO_SUCH_METHOD', 'ERROR_NO_SUCH_CAPCHA_ID', 'ERROR_TASK_ABSENT',
                            'ERROR_TASK_NOT_SUPPORTED', 'ERROR_FUNCAPTCHA_NOT_ALLOWED'):
            raise exceptions.MalformedRequestError(error_msg)
        elif error_code in ('ERROR_ZERO_CAPTCHA_FILESIZE', 'ERROR_TOO_BIG_CAPTCHA_FILESIZE',
                            'ERROR_WRONG_FILE_EXTENSION', 'ERROR_IMAGE_TYPE_NOT_SUPPORTED',
                            'ERROR_UPLOAD', 'ERROR_PAGEURL', 'ERROR_BAD_TOKEN_OR_PAGEURL',
                            'ERROR_GOOGLEKEY', 'ERROR_EMPTY_COMMENT',
                            'ERROR_INCORRECT_SESSION_DATA', 'ERROR_RECAPTCHA_INVALID_SITEKEY',
                            'ERROR_RECAPTCHA_INVALID_DOMAIN', 'ERROR_RECAPTCHA_OLD_BROWSER',
                            'ERROR_TOKEN_EXPIRED', 'ERROR_INVISIBLE_RECAPTCHA'):
            raise exceptions.BadInputDataError(error_msg)
        elif error_code in ('ERROR_CAPTCHAIMAGE_BLOCKED', 'ERROR_CAPTCHA_UNSOLVABLE',
                            'ERROR_BAD_DUPLICATES', 'ERROR_RECAPTCHA_TIMEOUT',
                            'ERROR_FAILED_LOADING_WIDGET'):
            raise exceptions.UnableToSolveError(error_msg)
        elif error_code in ('ERROR_PROXY_CONNECT_REFUSED', 'ERROR_PROXY_CONNECT_TIMEOUT',
                            'ERROR_PROXY_READ_TIMEOUT', 'ERROR_PROXY_BANNED',
                            'ERROR_PROXY_TRANSPARENT', 'ERROR_PROXY_HAS_NO_IMAGE_SUPPORT',
                            'ERROR_PROXY_INCOMPATIBLE_HTTP_VERSION', 'ERROR_PROXY_NOT_AUTHORISED'):
            raise exceptions.ProxyError(error_msg)

        raise exceptions.ServiceError(error_msg)


class GetBalanceRequest(Request):
    """ GetBalance Request class """

    def prepare(self) -> dict:   # type: ignore
        """ Prepares request """

        request = super().prepare()
        request.update(dict(url=self._service.BASE_URL + "/getBalance"))

        return request

    def parse_response(self, response) -> dict:
        """ Parses response and returns task_id """

        return dict(balance=float(super().parse_response(response)['balance']))


class GetStatusRequest(GetBalanceRequest):
    """ GetStatus Request class """

    def parse_response(self, response) -> dict:
        """ Parses response and returns task_id """

        try:
            return super().parse_response(response)
        except exceptions.UnicapsException:
            return {}


class ReportGoodRequest(Request):
    """ ReportGood Request class """

    # pylint: disable=arguments-differ
    def prepare(self, solved_captcha) -> dict:  # type: ignore
        """ Prepares request """

        raise exceptions.UnicapsException(
            "Report for good CAPTCHA is not supported by the current service!"
        )


class ReportBadRequest(Request):
    """ ReportBad Request class """

    # pylint: disable=arguments-differ
    def prepare(self, solved_captcha) -> dict:  # type: ignore
        """ Prepares request """

        request = super().prepare(solved_captcha=solved_captcha)

        captcha_type = solved_captcha.task.captcha.get_type()

        if captcha_type == CaptchaType.IMAGE:
            uri = "/reportIncorrectImageCaptcha"
        elif captcha_type in (CaptchaType.RECAPTCHAV2, CaptchaType.RECAPTCHAV3):
            uri = "/reportIncorrectRecaptcha"
        else:
            raise exceptions.UnicapsException(
                f"Report for bad {captcha_type.value} is not supported!"
            )

        request.update(dict(url=self._service.BASE_URL + uri))
        request["json"].update(dict(taskId=int(solved_captcha.captcha_id)))
        return request


class TaskRequest(Request):
    """ Request class for requests to /createTask """

    # pylint: disable=arguments-differ,unused-argument
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare a request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        request.update(dict(url=self._service.BASE_URL + "/createTask"))
        request["json"].update(
            dict(task={},
                 softId=940)
        )

        # add proxy
        if proxy:
            request["json"]["task"].update(
                dict(
                    proxyType=proxy.proxy_type.value,
                    # Anti-captcha supports IP addresses only
                    proxyAddress=proxy.get_ip_address(),
                    proxyPort=proxy.port
                )
            )

            if proxy.login:
                request["json"]["task"].update(
                    dict(
                        proxyLogin=proxy.login,
                        proxyPassword=proxy.password
                    )
                )

        if user_agent:
            request["json"]["task"]["userAgent"] = user_agent

        if cookies:
            request["json"]["task"]["cookies"] = '; '.join(f'{k}={v}'for k, v in cookies.items())

        return request

    def parse_response(self, response) -> dict:
        """ Parses response and returns task_id """

        response_data = super().parse_response(response)

        return {"task_id": response_data.pop("taskId"),
                "extra": response_data}


class SolutionRequest(Request):
    """ Request class for requests to /getTaskResult """

    # pylint: disable=arguments-differ
    def prepare(self, task) -> dict:  # type: ignore
        """ Prepare a request """

        request = super().prepare(task=task)
        request.update(dict(url=self._service.BASE_URL + "/getTaskResult"))
        request["json"].update(dict(taskId=str(task.task_id)))

        return request

    def parse_response(self, response) -> dict:
        """ Parses response and returns solution and cost """

        response_data = super().parse_response(response)

        if response_data["status"] != "ready":
            raise exceptions.SolutionNotReadyYet()

        solution_data = response_data["solution"]
        solution_class = self.source_data['task'].captcha.get_solution_class()
        captcha_type = self.source_data['task'].captcha.get_type()
        args = []
        kwargs = {}
        if captcha_type in (CaptchaType.IMAGE,):
            args.append(solution_data.pop('text'))
        elif captcha_type in (CaptchaType.RECAPTCHAV2, CaptchaType.RECAPTCHAV3,
                              CaptchaType.HCAPTCHA):
            args.append(solution_data.pop('gRecaptchaResponse'))
        elif captcha_type in (CaptchaType.FUNCAPTCHA,):
            args.append(solution_data.pop('token'))
        elif captcha_type in (CaptchaType.GEETEST, CaptchaType.GEETESTV4):
            kwargs.update(solution_data)
        else:
            kwargs.update(solution_data)

        solution = solution_class(*args, **kwargs)

        if "cost" in response_data:
            cost = response_data.pop("cost")
        else:
            cost = None

        return dict(
            solution=solution,
            cost=cost,
            extra=response_data
        )


class ImageCaptchaTaskRequest(TaskRequest):
    """ ImageCaptchaTask Request class """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare a request """

        request = super().prepare(
            captcha=captcha,
            proxy=None,
            user_agent=None,
            cookies=None
        )

        task_data = dict(
            type="ImageToTextTask",
            body=captcha.get_image_base64().decode('ascii')
        )
        task_data.update(
            captcha.get_optional_data(
                is_case_sensitive=('case', None),
                is_phrase=('phrase', None),
                is_math=('math', None),
                char_type=('numeric', lambda v: v.value if v.value in (1, 2) else None),
                min_len=('minLength', None),
                max_len=('maxLength', None),
                comment=('comment', None)
            )
        )
        request['json']['task'].update(task_data)

        # set workers pool language
        if captcha.language:
            request['json']['languagePool'] = (
                'rn' if captcha.language == WorkerLanguage.RUSSIAN else 'en'
            )

        return request


class ImageCaptchaSolutionRequest(SolutionRequest):
    """ Image CAPTCHA solution request """


class RecaptchaV2TaskRequest(TaskRequest):
    """ reCAPTCHA v2 task Request class """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepares request """

        if proxy:
            kwargs = dict(captcha=captcha, proxy=proxy, user_agent=user_agent, cookies=cookies)
            task_type = "RecaptchaV2EnterpriseTask" if captcha.is_enterprise else "NoCaptchaTask"
        else:
            kwargs = dict(captcha=captcha, proxy=None, user_agent=None, cookies=None)
            task_type = ("RecaptchaV2EnterpriseTaskProxyless" if captcha.is_enterprise
                         else "NoCaptchaTaskProxyless")

        request = super().prepare(**kwargs)
        request['json']['task'].update(
            dict(
                type=task_type,
                websiteURL=captcha.page_url,
                websiteKey=captcha.site_key,
                isInvisible=captcha.is_invisible
            )
        )

        # if enterprise captcha
        if captcha.is_enterprise:
            request['json']['task']['enterprisePayload'] = dict(s=captcha.data_s)
        else:
            # set optional data if any
            request['json']['task'].update(
                captcha.get_optional_data(
                    data_s=('recaptchaDataSValue', None)
                )
            )

        # set optional api_domain if any
        request['json']['task'].update(
            captcha.get_optional_data(
                api_domain=('apiDomain', None)
            )
        )

        return request


class RecaptchaV2SolutionRequest(SolutionRequest):
    """ reCAPTCHA v2 solution request """


class RecaptchaV3TaskRequest(TaskRequest):
    """ reCAPTCHA v3 task Request class """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepares request """

        request = super().prepare(
            captcha=captcha,
            proxy=None,
            user_agent=None,
            cookies=None
        )

        request['json']['task'].update(
            dict(
                type="RecaptchaV3TaskProxyless",
                websiteURL=captcha.page_url,
                websiteKey=captcha.site_key
            )
        )
        # set optional data if any
        request['json']['task'].update(
            captcha.get_optional_data(
                min_score=('minScore', None),
                action=('pageAction', None),
                api_domain=('apiDomain', None)
            )
        )

        # if enterprise captcha
        if captcha.is_enterprise:
            request['json']['task']['isEnterprise'] = True

        return request


class RecaptchaV3SolutionRequest(SolutionRequest):
    """ reCAPTCHA v3 solution request """


class FunCaptchaTaskRequest(TaskRequest):
    """ FunCaptcha task Request class """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepares request """

        if proxy:
            kwargs = dict(captcha=captcha, proxy=proxy, user_agent=user_agent, cookies=cookies)
            task_type = "FunCaptchaTask"
        else:
            kwargs = dict(captcha=captcha, proxy=None, user_agent=None, cookies=None)
            task_type = "FunCaptchaTaskProxyless"

        request = super().prepare(**kwargs)
        request['json']['task'].update(
            dict(
                type=task_type,
                websiteURL=captcha.page_url,
                websitePublicKey=captcha.public_key
            )
        )
        # set optional data if any
        request['json']['task'].update(
            captcha.get_optional_data(
                service_url=('funcaptchaApiJSSubdomain', None),
            )
        )

        # add blob value
        if captcha.blob:
            request['json']['task']['data'] = json.dumps(dict(blob=captcha.blob))

        return request


class FunCaptchaSolutionRequest(SolutionRequest):
    """ FunCaptcha solution request """


class GeeTestTaskRequest(TaskRequest):
    """ GeeTest task Request class """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepares request """

        if proxy:
            kwargs = dict(captcha=captcha, proxy=proxy, user_agent=user_agent, cookies=cookies)
            task_type = "GeeTestTask"
        else:
            kwargs = dict(captcha=captcha, proxy=None, user_agent=None, cookies=None)
            task_type = "GeeTestTaskProxyless"

        request = super().prepare(**kwargs)
        request['json']['task'].update(
            dict(
                type=task_type,
                websiteURL=captcha.page_url,
                gt=captcha.gt_key,
                challenge=captcha.challenge
            )
        )

        # set optional data if any
        request['json']['task'].update(
            captcha.get_optional_data(
                api_server=('geetestApiServerSubdomain', None),
            )
        )

        return request


class GeeTestSolutionRequest(SolutionRequest):
    """ GeeTest solution request """


class GeeTestV4TaskRequest(TaskRequest):
    """ GeeTest task Request class """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepares request """

        if proxy:
            kwargs = dict(captcha=captcha, proxy=proxy, user_agent=user_agent, cookies=cookies)
            task_type = "GeeTestTask"
        else:
            kwargs = dict(captcha=captcha, proxy=None, user_agent=None, cookies=None)
            task_type = "GeeTestTaskProxyless"

        request = super().prepare(**kwargs)
        request['json']['task'].update(
            dict(
                type=task_type,
                websiteURL=captcha.page_url,
                gt=captcha.captcha_id,
                version=4
            )
        )

        return request


class GeeTestV4SolutionRequest(SolutionRequest):
    """ GeeTest solution request """


class HCaptchaTaskRequest(TaskRequest):
    """ hCaptcha task Request class """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepares request """

        if proxy:
            kwargs = dict(captcha=captcha, proxy=proxy, user_agent=user_agent, cookies=cookies)
            task_type = "HCaptchaTask"
        else:
            kwargs = dict(captcha=captcha, proxy=None, user_agent=user_agent, cookies=None)
            task_type = "HCaptchaTaskProxyless"

        request = super().prepare(**kwargs)
        request['json']['task'].update(
            dict(
                type=task_type,
                websiteURL=captcha.page_url,
                websiteKey=captcha.site_key,
                isInvisible=captcha.is_invisible
            )
        )

        return request


class HCaptchaSolutionRequest(SolutionRequest):
    """ hCaptcha solution request """



# -*- coding: UTF-8 -*-

""" CAPTCHAs """

# pylint: disable=unused-import,import-error
from .image import ImageCaptcha
from .text import TextCaptcha
from .recaptcha_v2 import RecaptchaV2
from .recaptcha_v3 import RecaptchaV3
from .hcaptcha import HCaptcha
from .funcaptcha import FunCaptcha
from .keycaptcha import KeyCaptcha
from .geetest import GeeTest
from .geetest_v4 import GeeTestV4
from .capy import CapyPuzzle
from .tiktok import TikTokCaptcha
from .base import CaptchaType

__all__ = (
    'ImageCaptcha',
    'TextCaptcha',
    'RecaptchaV2',
    'RecaptchaV3',
    'HCaptcha',
    'FunCaptcha',
    'KeyCaptcha',
    'GeeTest',
    'GeeTestV4',
    'CapyPuzzle',
    'TikTokCaptcha',
    'CaptchaType'
)



# -*- coding: UTF-8 -*-
"""
2captcha.com service
"""

from .base import HTTPService
from .._transport.http_transport import HTTPRequestJSON  # type: ignore
from .. import exceptions
from .._captcha import CaptchaType
from ..common import CaptchaAlphabet

__all__ = [
    'Service', 'GetBalanceRequest', 'GetStatusRequest',
    'ReportGoodRequest', 'ReportBadRequest',
    'ImageCaptchaTaskRequest', 'ImageCaptchaSolutionRequest',
    'RecaptchaV2TaskRequest', 'RecaptchaV2SolutionRequest',
    'RecaptchaV3TaskRequest', 'RecaptchaV3SolutionRequest',
    'TextCaptchaTaskRequest', 'TextCaptchaSolutionRequest',
    'FunCaptchaTaskRequest', 'FunCaptchaSolutionRequest',
    'KeyCaptchaTaskRequest', 'KeyCaptchaSolutionRequest',
    'GeeTestTaskRequest', 'GeeTestSolutionRequest',
    'GeeTestV4TaskRequest', 'GeeTestV4SolutionRequest',
    'HCaptchaTaskRequest', 'HCaptchaSolutionRequest',
    'CapyPuzzleTaskRequest', 'CapyPuzzleSolutionRequest',
    'TikTokCaptchaTaskRequest', 'TikTokCaptchaSolutionRequest'
]


class Service(HTTPService):
    """ Main service class for 2captcha """

    BASE_URL = 'https://2captcha.com'

    def _post_init(self):
        """ Init settings """

        for captcha_type in self.settings:
            self.settings[captcha_type].polling_delay = 5
            self.settings[captcha_type].polling_interval = 5
            self.settings[captcha_type].solution_timeout = 180

            if captcha_type in (CaptchaType.RECAPTCHAV2, CaptchaType.HCAPTCHA):
                self.settings[captcha_type].polling_delay = 20
                self.settings[captcha_type].solution_timeout = 300
            elif captcha_type in (CaptchaType.RECAPTCHAV3,):
                self.settings[captcha_type].polling_delay = 15
            elif captcha_type in (CaptchaType.TIKTOK,):
                self.settings[captcha_type].polling_interval = 1


class Request(HTTPRequestJSON):
    """ Common Request class for 2captcha """

    def parse_response(self, response) -> dict:
        """ Parse response and checks for errors """

        response_data = super().parse_response(response)

        if response_data.pop("status") == 1:
            return response_data

        ###############
        # handle errors
        ###############
        error_code = response_data["request"]
        error_text = response_data.get("error_text", "")
        error_msg = f"{error_code}: {error_text}"

        if error_code == 'CAPCHA_NOT_READY':  # pylint: disable=no-else-raise
            raise exceptions.SolutionNotReadyYet()
        elif error_code in ('ERROR_WRONG_USER_KEY', 'ERROR_KEY_DOES_NOT_EXIST',
                            'ERROR_IP_NOT_ALLOWED', 'IP_BANNED'):
            raise exceptions.AccessDeniedError(error_msg)
        elif error_code in ('ERROR_ZERO_BALANCE',):
            raise exceptions.LowBalanceError(error_msg)
        elif error_code in ('ERROR_NO_SLOT_AVAILABLE',):
            # If server returns ERROR_NO_SLOT_AVAILABLE make a 5 seconds timeout before sending
            # next request.
            # time.sleep(5)
            raise exceptions.ServiceTooBusy(error_msg)
        elif error_code in ('MAX_USER_TURN',) or error_code.startswith('ERROR:'):
            raise exceptions.TooManyRequestsError(error_msg)
        elif error_code in ('ERROR_WRONG_ID_FORMAT', 'ERROR_WRONG_CAPTCHA_ID'):
            raise exceptions.MalformedRequestError(error_msg)
        elif error_code in ('ERROR_ZERO_CAPTCHA_FILESIZE', 'ERROR_TOO_BIG_CAPTCHA_FILESIZE',
                            'ERROR_WRONG_FILE_EXTENSION', 'ERROR_IMAGE_TYPE_NOT_SUPPORTED',
                            'ERROR_UPLOAD', 'ERROR_PAGEURL', 'ERROR_BAD_TOKEN_OR_PAGEURL',
                            'ERROR_GOOGLEKEY', 'ERROR_BAD_PARAMETERS', 'ERROR_TOKEN_EXPIRED',
                            'ERROR_EMPTY_ACTION'):
            raise exceptions.BadInputDataError(error_msg)
        elif error_code in ('ERROR_CAPTCHAIMAGE_BLOCKED', 'ERROR_CAPTCHA_UNSOLVABLE',
                            'ERROR_BAD_DUPLICATES'):
            raise exceptions.UnableToSolveError(error_msg)
        elif error_code in ('ERROR_BAD_PROXY', 'ERROR_PROXY_CONNECTION_FAILED'):
            raise exceptions.ProxyError(error_msg)

        raise exceptions.ServiceError(error_msg)


class InRequest(Request):
    """ Request class for requests to /in.php """

    def prepare(self, **kwargs) -> dict:
        """ Prepare request """

        request = super().prepare(**kwargs)
        request.update(
            dict(
                method="POST",
                url=self._service.BASE_URL + "/in.php",
                data=dict(
                    key=self._service.api_key,
                    json=1,
                    soft_id=2738
                )
            )
        )
        return request


class ResRequest(Request):
    """ Request class for requests to /res.php """

    def prepare(self, **kwargs) -> dict:
        """ Prepare request """

        request = super().prepare(**kwargs)
        request.update(
            dict(
                method="GET",
                url=self._service.BASE_URL + "/res.php",
                params=dict(
                    key=self._service.api_key,
                    json=1
                )
            )
        )
        return request


class GetBalanceRequest(ResRequest):
    """ GetBalance Request class """

    def prepare(self) -> dict:
        """ Prepare request """

        request = super().prepare()
        request["params"].update(dict(action="getbalance"))
        return request

    def parse_response(self, response) -> dict:
        """ Parse response and return balance """

        return {'balance': float(super().parse_response(response)["request"])}


class GetStatusRequest(GetBalanceRequest):
    """ GetStatus Request class """

    def parse_response(self, response) -> dict:
        """ Parse response and return status """

        try:
            return super().parse_response(response)
        except exceptions.UnicapsException:
            return {}


class ReportGoodRequest(ResRequest):
    """ ReportGood Request class """

    # pylint: disable=arguments-differ
    def prepare(self, solved_captcha) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(solved_captcha=solved_captcha)
        request["params"].update(
            dict(
                action="reportgood",
                id=solved_captcha.captcha_id
            )
        )
        return request


class ReportBadRequest(ResRequest):
    """ ReportBad Request class """

    # pylint: disable=arguments-differ
    def prepare(self, solved_captcha) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(solved_captcha=solved_captcha)
        request["params"].update(
            dict(
                action="reportbad",
                id=solved_captcha.captcha_id
            )
        )
        return request


class TaskRequest(InRequest):
    """ Common Task Request class """

    # pylint: disable=arguments-differ,unused-argument
    def prepare(self, captcha, proxy, user_agent, cookies):
        """ Prepare a request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        if proxy:
            request['data'].update(
                dict(
                    proxy=proxy.get_string(),
                    proxytype=proxy.proxy_type.value.upper()
                )
            )

        if cookies:
            request['data']['cookies'] = ';'.join([f'{k}:{v}' for k, v in cookies.items()])

        if user_agent:
            request['data']['userAgent'] = user_agent

        return request

    def parse_response(self, response) -> dict:
        """ Parse response and return task_id """

        response_data = super().parse_response(response)

        return dict(
            task_id=response_data.pop("request"),
            extra=response_data
        )


class SolutionRequest(ResRequest):
    """ Common Solution Request class """

    # pylint: disable=arguments-differ
    def prepare(self, task) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(task=task)
        request["params"].update(
            dict(action="get2", id=task.task_id)
        )
        return request

    def parse_response(self, response) -> dict:
        """ Parse response and return solution and cost """

        response_data = super().parse_response(response)

        solution_data = response_data.pop("request")
        solution_class = self.source_data['task'].captcha.get_solution_class()
        captcha_type = self.source_data['task'].captcha.get_type()
        if captcha_type == CaptchaType.GEETEST:
            solution = solution_class(
                challenge=solution_data['geetest_challenge'],
                validate=solution_data['geetest_validate'],
                seccode=solution_data['geetest_seccode']
            )
        elif captcha_type == CaptchaType.GEETESTV4:
            solution = solution_class(
                captcha_id=solution_data['captcha_id'],
                lot_number=solution_data['lot_number'],
                pass_token=solution_data['pass_token'],
                gen_time=solution_data['gen_time'],
                captcha_output=solution_data['captcha_output'],
            )
        elif captcha_type == CaptchaType.CAPY:
            solution = solution_class(**solution_data)
        elif captcha_type in (CaptchaType.TIKTOK,):
            solution = solution_class(
                dict(
                    [key_value.split(':', maxsplit=1) for key_value in solution_data.split(';')]
                )
            )
        else:
            solution = solution_class(solution_data)

        price = None
        if 'price' in response_data:
            price = response_data.pop("price")

        return dict(
            solution=solution,
            cost=price,
            extra=response_data
        )


class ImageCaptchaTaskRequest(TaskRequest):
    """ ImageCaptchaTask Request class """

    # pylint: disable=arguments-differ,unused-argument,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        # add required params
        request['data'].update(
            dict(
                method="base64",
                body=captcha.get_image_base64().decode('ascii')
            )
        )

        # add optional params
        request['data'].update(
            captcha.get_optional_data(
                is_phrase=('phrase', lambda v: int(bool(v))),
                is_case_sensitive=('regsense', lambda v: int(bool(v))),
                char_type=('numeric', lambda v: v.value),
                is_math=('calc', lambda v: int(bool(v))),
                min_len=('min_len', None),
                max_len=('max_len', None),
                alphabet=('language',
                          lambda v: {CaptchaAlphabet.CYRILLIC: 1,
                                     CaptchaAlphabet.LATIN: 2}.get(v, 0)),
                language=('lang', lambda v: v.value),
                comment=('textinstructions', None),
            )
        )

        return request


class ImageCaptchaSolutionRequest(SolutionRequest):
    """ Image CAPTCHA solution request """


class RecaptchaV2TaskRequest(TaskRequest):
    """ reCAPTCHA v2 task request """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )
        request['data'].update(
            dict(
                method="userrecaptcha",
                googlekey=captcha.site_key,
                pageurl=captcha.page_url,
                invisible=int(captcha.is_invisible)
            )
        )

        # add optional params
        request['data'].update(
            captcha.get_optional_data(
                data_s=('data-s', None),
                api_domain=('domain', None)
            )
        )

        # check if enterprise captcha
        if captcha.is_enterprise:
            request['data']['enterprise'] = 1

        return request


class RecaptchaV2SolutionRequest(SolutionRequest):
    """ reCAPTCHA v2 solution request """


class RecaptchaV3TaskRequest(TaskRequest):
    """ reCAPTCHA v3 task request """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )
        request['data'].update(
            dict(
                method="userrecaptcha",
                version="v3",
                googlekey=captcha.site_key,
                pageurl=captcha.page_url
            )
        )

        # add optional params
        request['data'].update(
            captcha.get_optional_data(
                action=('action', None),
                min_score=('min_score', None),
                api_domain=('domain', None)
            )
        )

        # check if enterprise captcha
        if captcha.is_enterprise:
            request['data']['enterprise'] = 1

        return request


class RecaptchaV3SolutionRequest(SolutionRequest):
    """ reCAPTCHA v3 solution request """


class TextCaptchaTaskRequest(TaskRequest):
    """ TextCaptcha task request """

    # pylint: disable=arguments-differ,unused-argument,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        request['data'].update(
            dict(
                textcaptcha=captcha.text
            )
        )

        # add optional params
        request['data'].update(
            captcha.get_optional_data(
                alphabet=('language',
                          lambda v: {CaptchaAlphabet.CYRILLIC: 1,
                                     CaptchaAlphabet.LATIN: 2}.get(v, 0)),
                language=('lang', lambda v: v.value)
            )
        )

        return request


class TextCaptchaSolutionRequest(SolutionRequest):
    """ TextCaptcha solution request """


class FunCaptchaTaskRequest(TaskRequest):
    """ FunCaptcha task request """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        request['data'].update(
            dict(
                method="funcaptcha",
                publickey=captcha.public_key,
                pageurl=captcha.page_url
            )
        )

        # add optional params
        request['data'].update(
            captcha.get_optional_data(
                service_url=('surl', None),
                no_js=('nojs', None),
                blob=('data[blob]', None),
            )
        )

        return request


class FunCaptchaSolutionRequest(SolutionRequest):
    """ FunCaptcha solution request """


class KeyCaptchaTaskRequest(TaskRequest):
    """ KeyCaptcha task request """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        request['data'].update(
            dict(
                method="keycaptcha",
                s_s_c_user_id=captcha.user_id,
                s_s_c_session_id=captcha.session_id,
                s_s_c_web_server_sign=captcha.ws_sign,
                s_s_c_web_server_sign2=captcha.ws_sign2,
                pageurl=captcha.page_url
            )
        )

        return request


class KeyCaptchaSolutionRequest(SolutionRequest):
    """ KeyCaptcha solution request """


class GeeTestTaskRequest(TaskRequest):
    """ GeeTest task request """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        request['data'].update(
            dict(
                method="geetest",
                gt=captcha.gt_key,
                challenge=captcha.challenge,
                pageurl=captcha.page_url
            )
        )

        # add optional params
        request['data'].update(
            captcha.get_optional_data(
                api_server=('api_server', None)
            )
        )

        return request


class GeeTestSolutionRequest(SolutionRequest):
    """ GeeTest solution request """


class HCaptchaTaskRequest(TaskRequest):
    """ HCaptcha task request """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        request['data'].update(
            dict(
                method="hcaptcha",
                sitekey=captcha.site_key,
                pageurl=captcha.page_url,
                invisible=int(captcha.is_invisible)
            )
        )

        # add optional params
        request['data'].update(
            captcha.get_optional_data(
                api_domain=('domain', None)
            )
        )

        return request


class HCaptchaSolutionRequest(SolutionRequest):
    """ HCaptcha solution request """


class CapyPuzzleTaskRequest(TaskRequest):
    """ CapyPuzzle task request """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        request['data'].update(
            dict(
                method="capy",
                captchakey=captcha.site_key,
                pageurl=captcha.page_url
            )
        )

        # add optional params
        request['data'].update(
            captcha.get_optional_data(
                api_server=('api_server', None)
            )
        )

        return request


class CapyPuzzleSolutionRequest(SolutionRequest):
    """ CapyPuzzle solution request """


class TikTokCaptchaTaskRequest(TaskRequest):
    """ TikTokCaptcha task request """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        # default values for additional params
        # see https://2captcha.com/2captcha-api#solving_tiktok for details
        add_params_mapping = {
            'https://www.tiktok.com/login/phone-or-email/email': {
                'aid': 1459,
                'host': 'https://www-useast1a.tiktok.com'
            },
            'https://ads.tiktok.com/i18n/signup': {
                'aid': 1583,
                'host': 'https://verify-sg.byteoversea.com'
            },
            'default': {'aid': None, 'host': None}
        }

        add_params = add_params_mapping[next(
            (url for url in add_params_mapping if captcha.page_url.startswith(url)),
            'default'
        )]

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        request['data'].update(
            dict(
                method="tiktok",
                pageurl=captcha.page_url,
                aid=add_params['aid'] if captcha.aid is None else captcha.aid,
                host=add_params['host'] if captcha.host is None else captcha.host
            )
        )

        return request


class TikTokCaptchaSolutionRequest(SolutionRequest):
    """ TikTokCaptcha solution request """


class GeeTestV4TaskRequest(TaskRequest):
    """ GeeTest v4 task request """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        request['data'].update(
            dict(
                method="geetest_v4 ",
                captcha_id=captcha.captcha_id,
                pageurl=captcha.page_url
            )
        )

        return request


class GeeTestV4SolutionRequest(SolutionRequest):
    """ GeeTest v4 solution request """



# -*- coding: UTF-8 -*-
"""
Transport and requests for HTTP protocol
"""

from json.decoder import JSONDecodeError
from typing import Optional, Dict

import httpx

from .base import BaseTransport, BaseRequest  # type: ignore
from ..exceptions import NetworkError, ServiceError  # type: ignore
from ..__version__ import __version__  # type: ignore


HTTP_RETRY_MAX_COUNT = 5  # max retry count in case of http(s) errors
HTTP_RETRY_BACKOFF_FACTOR = 0.5  # backoff factor for Retry
HTTP_RETRY_STATUS_FORCELIST = {500, 502, 503, 504}  # status forcelist for Retry


class StandardHTTPTransport(BaseTransport):  # pylint: disable=too-few-public-methods
    """ Standard HTTP Transport """

    def __init__(self, settings: Optional[Dict] = None):
        super().__init__(settings)
        self.settings.setdefault('max_retries', HTTP_RETRY_MAX_COUNT)
        self.settings.setdefault('handle_http_errors', True)

        default_headers = {
            'User-Agent': f'python-unicaps/{__version__}'
        }

        self.session = httpx.Client(
            headers=default_headers,
            timeout=httpx.Timeout(timeout=30)
        )
        self.session_async = httpx.AsyncClient(
            headers=default_headers,
            timeout=httpx.Timeout(timeout=30)
        )

    def _make_request(self, request_data: Dict) -> httpx.Response:
        if 'headers' not in request_data:
            request_data['headers'] = {}

        try:
            response = self.session.request(**request_data)
        except httpx.TimeoutException as exc:
            raise NetworkError('Timeout') from exc
        except httpx.RequestError as exc:
            raise NetworkError('RequestError') from exc

        if self.settings['handle_http_errors']:
            try:
                response.raise_for_status()
            except httpx.HTTPStatusError as exc:
                raise NetworkError('HTTPStatusError') from exc

        return response

    async def _make_request_async(self, request_data: Dict) -> httpx.Response:
        if 'headers' not in request_data:
            request_data['headers'] = {}

        try:
            response = await self.session_async.request(**request_data)
        except httpx.TimeoutException as exc:
            raise NetworkError('Timeout') from exc
        except httpx.RequestError as exc:
            raise NetworkError('RequestError') from exc

        if self.settings['handle_http_errors']:
            try:
                response.raise_for_status()
            except httpx.HTTPStatusError as exc:
                raise NetworkError('HTTPStatusError') from exc

        return response

    def close(self):
        """ Close connections """
        self.session.close()

    async def close_async(self):
        """ Close connections (async) """
        await self.session_async.aclose()


class HTTPRequestJSON(BaseRequest):
    """ HTTP Request that returns JSON response """

    def prepare(self, **kwargs) -> Dict:
        """ Prepares request """

        request = super().prepare(**kwargs)
        request.update(
            dict(headers={'Accept': 'application/json'})
        )
        return request

    def parse_response(self, response: httpx.Response) -> Dict:
        """ Parses response """

        try:
            return response.json()
        except JSONDecodeError as exc:
            raise ServiceError("Unable to parse response from the server: bad JSON") from exc



# -*- coding: UTF-8 -*-
"""
Text CAPTCHA
"""

from dataclasses import dataclass
from typing import Optional

from enforce_typing import enforce_types  # type: ignore

from .base import BaseCaptcha, BaseCaptchaSolution
from ..common import CaptchaAlphabet, WorkerLanguage


@enforce_types
@dataclass
class TextCaptcha(BaseCaptcha):
    """ Text CAPTCHA """

    text: str
    alphabet: Optional[CaptchaAlphabet] = None
    language: Optional[WorkerLanguage] = None


@enforce_types
@dataclass
class TextCaptchaSolution(BaseCaptchaSolution):
    """ Text CAPTCHA solution """

    text: str



# -*- coding: UTF-8 -*-
"""
CaptchaSolver tests
"""
from unittest.mock import Mock

import pytest
from unicaps import CaptchaSolver, CaptchaSolvingService
from unicaps.captcha import CaptchaType

API_KEY = 'TEST_API_KEY'


@pytest.fixture(scope="module")
def captcha_solver():
    return CaptchaSolver('2captcha.com', API_KEY)


@pytest.fixture()
def mocked_captcha_solver(captcha_solver, monkeypatch):
    monkeypatch.setattr(captcha_solver, '_service', Mock())
    return captcha_solver


def test_solver_init():
    service = CaptchaSolvingService.ANTI_CAPTCHA
    solver = CaptchaSolver(service, API_KEY)

    assert solver.service_name == service
    assert solver.api_key == API_KEY


def test_solver_init_from_string():
    solver = CaptchaSolver('2captcha.com', API_KEY)

    assert solver.service_name == CaptchaSolvingService.TWOCAPTCHA
    assert solver.api_key == API_KEY


def test_solver_bad_init():
    with pytest.raises(ValueError):
        CaptchaSolver('2captcha', API_KEY)


def test_solver_bad_init2():
    with pytest.raises(ValueError):
        CaptchaSolver(b'2captcha.com', API_KEY)


# @pytest.mark.parametrize("captcha_type", CaptchaType)
def test_call_solve_func(mocked_captcha_solver, captcha_instance):
    mapping = {
        CaptchaType.IMAGE: 'image_captcha',
        CaptchaType.TEXT: 'text_captcha',
        CaptchaType.RECAPTCHAV2: 'recaptcha_v2',
        CaptchaType.RECAPTCHAV3: 'recaptcha_v3',
        CaptchaType.HCAPTCHA: 'hcaptcha',
        CaptchaType.FUNCAPTCHA: 'funcaptcha',
        CaptchaType.KEYCAPTCHA: 'keycaptcha',
        CaptchaType.GEETEST: 'geetest',
        CaptchaType.CAPY: 'capy_puzzle',
        CaptchaType.TIKTOK: 'tiktok',
        CaptchaType.GEETESTV4: 'geetest_v4'
    }

    func = getattr(mocked_captcha_solver, f'solve_{mapping[captcha_instance.get_type()]}')
    func(
        **{k: v for k, v in captcha_instance.__dict__.items() if not k.startswith('_')}
    )



# -*- coding: UTF-8 -*-
"""
deathbycaptcha.com service
"""
import json

from .base import HTTPService
from .._transport.http_transport import HTTPRequestJSON  # type: ignore
from .. import exceptions
from .._captcha import CaptchaType

__all__ = [
    'Service', 'GetBalanceRequest', 'GetStatusRequest',
    'ReportGoodRequest', 'ReportBadRequest',
    'ImageCaptchaTaskRequest', 'ImageCaptchaSolutionRequest',
    'RecaptchaV2TaskRequest', 'RecaptchaV2SolutionRequest',
    'RecaptchaV3TaskRequest', 'RecaptchaV3SolutionRequest',
    'FunCaptchaTaskRequest', 'FunCaptchaSolutionRequest',
    'HCaptchaTaskRequest', 'HCaptchaSolutionRequest'
]


class Service(HTTPService):
    """ Main service class for deathbycaptcha """

    BASE_URL = 'http://api.dbcapi.me/api'

    def _post_init(self):
        """ Init settings """
        self._transport.settings['handle_http_errors'] = False

        for captcha_type in self.settings:
            self.settings[captcha_type].polling_delay = 5
            self.settings[captcha_type].polling_interval = 2
            self.settings[captcha_type].solution_timeout = 180

            if captcha_type in (CaptchaType.RECAPTCHAV2, CaptchaType.HCAPTCHA):
                self.settings[captcha_type].polling_delay = 15
                self.settings[captcha_type].solution_timeout = 200
            elif captcha_type in (CaptchaType.RECAPTCHAV3,):
                self.settings[captcha_type].polling_delay = 15


class Request(HTTPRequestJSON):
    """ Common Request class for deathbycaptcha """

    def prepare(self, **kwargs) -> dict:
        """ Prepare the request """

        request = super().prepare(**kwargs)

        method = kwargs.get('method', 'GET')
        data_or_params = 'data' if method == 'POST' else 'params'

        request.update({
            'method': kwargs.get('method', 'GET'),
            'url': self._service.BASE_URL + kwargs.get('url', ''),
            data_or_params: dict(
                authtoken=self._service.api_key
            )
        })
        return request

    def parse_response(self, response) -> dict:
        """ Parse response and check for errors """

        response_data = super().parse_response(response)

        status = response_data.get('status')
        if (response.status_code == 303 or response.is_success) and status == 0:
            response_data.pop('status')
            return response_data

        #################
        # handle errors #
        #################
        if response_data.get('error'):
            error_text = response_data['error']
        elif response.is_error:
            error_text = f'[{response.status_code} {response.reason_phrase}]'
        else:
            error_text = 'Unknown error'

        error_msg = f"{status}: {error_text}"

        if error_text in ('token authentication disabled', 'not-logged-in', 'banned'):
            raise exceptions.AccessDeniedError(error_msg)
        if error_text in ('insufficient-funds',):
            raise exceptions.LowBalanceError(error_msg)
        if error_text in ('service-overload',):
            raise exceptions.ServiceTooBusy(error_msg)
        if error_text in ('upload-failed', 'invalid-captcha'):
            raise exceptions.MalformedRequestError(error_msg)
        if error_text in ('ERROR_PAGEURL', 'Invalid base64-encoded CAPTCHA',
                          'Not a (CAPTCHA) image', 'Empty CAPTCHA image', 'ERROR_GOOGLEKEY',
                          'ERROR_PAGEURL', 'ERROR_PUBLICKEY', 'ERROR_SITEKEY', 'ERROR_ACTION',
                          'ERROR_MIN_SCORE', 'ERROR_MIN_SCORE_NOT_FLOAT'):
            raise exceptions.BadInputDataError(error_msg)
        if error_text in ('ERROR_PROXYTYPE', 'ERROR_PROXY'):
            raise exceptions.ProxyError(error_msg)

        raise exceptions.ServiceError(error_msg)


class PostRequest(Request):
    """ Request class for POST requests """

    def prepare(self, **kwargs) -> dict:
        """ Prepare request """
        return super().prepare(method="POST", **kwargs)


class GetRequest(Request):
    """ Request class for GET requests """

    def prepare(self, **kwargs) -> dict:
        """ Prepare request """
        return super().prepare(method="GET", **kwargs)


class GetBalanceRequest(GetRequest):
    """ GetBalance Request class """

    def prepare(self) -> dict:
        """ Prepare request """
        return super().prepare(url='')

    def parse_response(self, response) -> dict:
        """ Parse response and return balance """
        return {
            'balance': float(super().parse_response(response)["balance"]) / 100
        }


class GetStatusRequest(GetRequest):
    """ GetStatus Request class """

    def prepare(self):
        """ Prepare request """
        return super().prepare(url='/status')

    def parse_response(self, response) -> dict:
        """ Parse response and return status """
        try:
            response_data = super().parse_response(response)
            if response_data.get('is_service_overloaded'):
                return {}
            return response_data
        except exceptions.UnicapsException:
            return {}


class ReportGoodRequest(PostRequest):
    """ ReportGood Request class """

    # pylint: disable=arguments-differ
    def prepare(self, solved_captcha) -> dict:  # type: ignore
        """ Prepares request """
        raise exceptions.UnicapsException(
            "Report for good CAPTCHA is not supported by the current service!"
        )


class ReportBadRequest(PostRequest):
    """ ReportBad Request class """

    # pylint: disable=arguments-differ
    def prepare(self, solved_captcha) -> dict:  # type: ignore
        """ Prepare request """
        return super().prepare(
            url=f'/captcha/{solved_captcha.captcha_id}/report',
            solved_captcha=solved_captcha
        )


class TaskRequest(PostRequest):
    """ Common Task Request class """

    # pylint: disable=arguments-differ,unused-argument
    def prepare(self, captcha, proxy, user_agent, cookies):
        """ Prepare a request """

        request = super().prepare(
            url='/captcha',
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        return request

    def parse_response(self, response) -> dict:
        """ Parse response and return captcha_id """

        response_data = super().parse_response(response)

        # raise the BadInputDataError if CAPTCHA is not correct
        if not response_data.pop('is_correct'):
            raise exceptions.BadInputDataError('is_correct=false')

        if 'text' in response_data:
            response_data.pop('text')

        return dict(
            task_id=response_data.pop("captcha"),
            extra=response_data
        )


class SolutionRequest(GetRequest):
    """ Common Solution Request class """

    # pylint: disable=arguments-differ
    def prepare(self, task) -> dict:  # type: ignore
        """ Prepare request """
        return super().prepare(url=f'/captcha/{task.task_id}', task=task)

    def parse_response(self, response) -> dict:
        """ Parse response and return solution and cost """

        response_data = super().parse_response(response)

        # raise the UnableToSolveError if CAPTCHA is not correct
        if not response_data.pop('is_correct'):
            raise exceptions.UnableToSolveError('is_correct=false')

        # the empty text field means that solving is in progress
        text = response_data.pop("text")
        if not text:
            raise exceptions.SolutionNotReadyYet()

        # get solution class and prepare a solution object
        solution_class = self.source_data['task'].captcha.get_solution_class()
        solution = solution_class(text)

        response_data.pop("captcha")

        return dict(
            solution=solution,
            extra=response_data
        )


class ImageCaptchaTaskRequest(TaskRequest):
    """ ImageCaptchaTask Request class """

    # pylint: disable=arguments-differ,unused-argument,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        # add required params
        request['data'].update(dict(
            captchafile='base64:' + captcha.get_image_base64().decode('ascii')
        ))
        return request


class ImageCaptchaSolutionRequest(SolutionRequest):
    """ Image CAPTCHA solution request """


class RecaptchaV2TaskRequest(TaskRequest):
    """ reCAPTCHA v2 task request """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        data = {
            "googlekey": captcha.site_key,
            "pageurl": captcha.page_url
        }
        data.update(
            captcha.get_optional_data(
                data_s=('data-s', None),
            )
        )

        request['data'].update(
            dict(
                type=4,
                token_params=_dumps(data, proxy)
            )
        )

        return request


class RecaptchaV2SolutionRequest(SolutionRequest):
    """ reCAPTCHA v2 solution request """


class RecaptchaV3TaskRequest(TaskRequest):
    """ reCAPTCHA v3 task request """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        data = {
            "googlekey": captcha.site_key,
            "pageurl": captcha.page_url
        }

        data.update(
            captcha.get_optional_data(
                action=('action', None),
                min_score=('min_score', None),
            )
        )

        request['data'].update(
            dict(
                type=5,
                token_params=_dumps(data, proxy)
            )
        )

        return request


class RecaptchaV3SolutionRequest(SolutionRequest):
    """ reCAPTCHA v3 solution request """


class FunCaptchaTaskRequest(TaskRequest):
    """ FunCaptcha task request """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        data = {
            "publickey": captcha.public_key,
            "pageurl": captcha.page_url
        }

        request['data'].update(
            dict(
                type=6,
                funcaptcha_params=_dumps(data, proxy)
            )
        )

        return request


class FunCaptchaSolutionRequest(SolutionRequest):
    """ FunCaptcha solution request """


class HCaptchaTaskRequest(TaskRequest):
    """ HCaptcha task request """

    # pylint: disable=arguments-differ,signature-differs
    def prepare(self, captcha, proxy, user_agent, cookies) -> dict:  # type: ignore
        """ Prepare request """

        request = super().prepare(
            captcha=captcha,
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

        data = {
            "sitekey": captcha.site_key,
            "pageurl": captcha.page_url
        }

        request['data'].update(
            dict(
                type=7,
                hcaptcha_params=_dumps(data, proxy)
            )
        )

        return request


class HCaptchaSolutionRequest(SolutionRequest):
    """ HCaptcha solution request """


def _dumps(data, proxy):
    if proxy:
        data.update(
            dict(
                proxy=proxy.get_string(including_type=True),
                proxytype=proxy.proxy_type.value.upper()
            )
        )
    return json.dumps(data)



# -*- coding: UTF-8 -*-
"""
unicaps.exceptions
~~~~~~~~~~~~~~~~~~~

This module contains the set of Unicaps' exceptions.
"""


class UnicapsException(Exception):
    """Main exception class"""


class SolutionNotReadyYet(UnicapsException):
    """CAPTCHA solving in progress"""


class ServiceError(UnicapsException):
    """Main service-related exception class"""


class CaptchaError(UnicapsException):
    """CAPTCHA-related exception"""


class NetworkError(UnicapsException):
    """
    Network Connection Error
    Service returned 5xx status code
    """


class ProxyError(UnicapsException):
    """
    Bad proxy
    """


class AccessDeniedError(ServiceError):
    """
    Wrong API key
    IP banned
    IP not allowed
    """


class LowBalanceError(ServiceError):
    """
    Low balance
    """


class ServiceTooBusy(ServiceError):
    """
    No available slots
    """


class SolutionWaitTimeout(ServiceError):
    """
    Didn't receive solution within N minutes
    """


class TooManyRequestsError(ServiceError):
    """
    Exceeded request limit
    """


class MalformedRequestError(ServiceError):
    """
    Exceeded request limit
    """


class BadInputDataError(CaptchaError):
    """
    Not supported image file
    Empty file
    Image file is too big
    Bad captcha data (eg, wrong googlekey, bad page URL, etc.)
    """


class UnableToSolveError(CaptchaError):
    """
    Captcha unsolvable
    """



# -*- coding: UTF-8 -*-
"""
Unicaps package
~~~~~~~~~~~~~~~
"""

# pylint: disable=unused-import,import-error
from ._solver import CaptchaSolver
from ._solver_async import AsyncCaptchaSolver
from ._service import CaptchaSolvingService

__all__ = ('CaptchaSolver', 'AsyncCaptchaSolver', 'CaptchaSolvingService')



# -*- coding: UTF-8 -*-
"""
Supported CAPTCHAs
~~~~~~~~~~~~~~~~~~
"""

# pylint: disable=unused-import,import-error
from ._captcha import (ImageCaptcha, TextCaptcha, RecaptchaV2, RecaptchaV3, HCaptcha, FunCaptcha,
                       KeyCaptcha, GeeTest, GeeTestV4, CapyPuzzle, TikTokCaptcha, CaptchaType)

__all__ = (
    'ImageCaptcha',
    'TextCaptcha',
    'RecaptchaV2',
    'RecaptchaV3',
    'HCaptcha',
    'FunCaptcha',
    'KeyCaptcha',
    'GeeTest',
    'GeeTestV4',
    'CapyPuzzle',
    'TikTokCaptcha',
    'CaptchaType'
)



# -*- coding: UTF-8 -*-
"""
FunCaptcha
"""

from dataclasses import dataclass
from typing import Optional

from enforce_typing import enforce_types  # type: ignore

from .base import BaseCaptcha, BaseCaptchaSolution


@enforce_types
@dataclass
class FunCaptcha(BaseCaptcha):
    """ FunCaptcha """

    public_key: str
    page_url: str
    service_url: Optional[str] = None
    no_js: Optional[bool] = None
    blob: Optional[str] = None


@enforce_types
@dataclass
class FunCaptchaSolution(BaseCaptchaSolution):
    """ FunCaptcha solution """

    token: str



# -*- coding: UTF-8 -*-
"""
GeeTest CAPTCHA
"""

from dataclasses import dataclass
from typing import Optional

from enforce_typing import enforce_types  # type: ignore

from .base import BaseCaptcha, BaseCaptchaSolution


@enforce_types
@dataclass
class GeeTest(BaseCaptcha):
    """ GeeTest """

    page_url: str
    gt_key: str
    challenge: str
    api_server: Optional[str] = None


@enforce_types
@dataclass
class GeeTestSolution(BaseCaptchaSolution):
    """ GeeTest solution """

    challenge: str
    validate: str
    seccode: str



# -*- coding: UTF-8 -*-
"""
Base CAPTCHA stuff
"""

import enum
import importlib
from abc import ABC
from dataclasses import asdict, dataclass, fields, MISSING
from typing import Dict


class CaptchaType(enum.Enum):
    """ Captcha type enumeration """

    IMAGE = "ImageCaptcha"
    RECAPTCHAV2 = "RecaptchaV2"
    RECAPTCHAV3 = "RecaptchaV3"
    TEXT = "TextCaptcha"
    FUNCAPTCHA = "FunCaptcha"
    GEETEST = "GeeTest"
    GEETESTV4 = "GeeTestV4"
    HCAPTCHA = "HCaptcha"
    KEYCAPTCHA = "KeyCaptcha"
    CAPY = "CapyPuzzle"
    TIKTOK = "TikTokCaptcha"


@dataclass
class BaseCaptcha(ABC):
    """ Base class for any CAPTCHA """

    @classmethod
    def get_type(cls) -> CaptchaType:
        """ Return CaptchaType """

        return CaptchaType(cls.__name__)

    @classmethod
    def get_solution_class(cls) -> 'BaseCaptchaSolution':
        """ Return appropriate solution class """

        return getattr(importlib.import_module(cls.__module__), cls.__name__ + "Solution")

    def get_optional_data(self, **kwargs) -> Dict:
        """
        Return a dict with all optional fields requested (that are not None)
        as a dict with given names.

        :return: :dict:Dictionary of optional not None fields with given names and those values
        :rtype: dict
        """

        result = {}

        if not kwargs:
            # get all optional params
            kwargs = {
                field.name: (field.name, None) for field in fields(self)
                if field.default is not MISSING
            }

        for opt_field in kwargs:
            opt_field_value = getattr(self, opt_field)
            field_name, converter = kwargs[opt_field]
            if opt_field_value is not None:
                if callable(converter):
                    opt_field_value = converter(opt_field_value)
                result[field_name] = opt_field_value
        return result


@dataclass
class BaseCaptchaSolution(ABC):
    """ Base class for any CAPTCHA solution """

    @classmethod
    def get_type(cls) -> CaptchaType:
        """ Returns CaptchaType """

        return CaptchaType(cls.__name__.split("Solution", maxsplit=1)[0])

    @classmethod
    def get_captcha_class(cls) -> BaseCaptcha:
        """ Returns appropriate solution class """

        return getattr(
            importlib.import_module(cls.__module__),
            cls.__name__.split("Solution", maxsplit=1)[0]
        )

    def __str__(self):
        return '\n'.join(str(getattr(self, field.name)) for field in fields(self))

    def as_dict(self):
        """ Get solution data as Python dictionary """
        return asdict(self)



# -*- coding: UTF-8 -*-
"""
Proxy Server representation
"""
import socket

from dataclasses import dataclass
from enum import Enum
from typing import Optional


def _is_ip_address(value):
    try:
        socket.inet_aton(value)
        return True
    except socket.error:
        return False


class ProxyServerType(Enum):
    """ Type of proxy server """

    HTTP = 'http'
    HTTPS = 'https'
    SOCKS4 = 'socks4'
    SOCKS5 = 'socks5'


@dataclass
class ProxyServer:
    """ Represents Proxy server """

    address: str
    proxy_type: ProxyServerType = ProxyServerType.HTTP
    port: int = 80
    login: Optional[str] = None
    password: Optional[str] = None

    def __post_init__(self):
        proxy_string = self.address
        if '://' in proxy_string:
            proxy_type, proxy_string = proxy_string.split('://')
            self.proxy_type = ProxyServerType(proxy_type.lower())

        if '@' in proxy_string:
            credentials, proxy_string = proxy_string.split('@')
            self.login, self.password = credentials.split(':', maxsplit=1)

        if ':' in proxy_string:
            self.address, port = proxy_string.split(':', maxsplit=1)
            self.port = int(port)
        else:
            self.address = proxy_string

    def __str__(self):
        return self.get_string(including_type=True)

    def get_string(self, including_type=False):
        """ Get proxy as string like [<type>://][<login>:<password>@]<addr>:<port> """

        proxy_string = ''
        if including_type:
            proxy_string += self.proxy_type.value + '://'

        if self.login:
            proxy_string += self.login + ':' + self.password + '@'

        return proxy_string + self.address + ':' + str(self.port)

    def get_ip_address(self):
        """ Get IP address by hostname """

        if not _is_ip_address(self.address):
            return socket.gethostbyname(self.address)
        return self.address



# -*- coding: UTF-8 -*-
"""
Image CAPTCHA
"""

import base64
import imghdr
import io
import pathlib
from dataclasses import dataclass
from typing import Union, Optional

from enforce_typing import enforce_types  # type: ignore

from .base import BaseCaptcha, BaseCaptchaSolution
from ..common import CaptchaAlphabet, CaptchaCharType, WorkerLanguage
from ..exceptions import BadInputDataError


@enforce_types
@dataclass
class ImageCaptcha(BaseCaptcha):
    """ Image CAPTCHA """

    image: Union[bytes, io.RawIOBase, io.BufferedIOBase, pathlib.Path]
    char_type: Optional[CaptchaCharType] = None
    is_phrase: Optional[bool] = None
    is_case_sensitive: Optional[bool] = None
    is_math: Optional[bool] = None
    min_len: Optional[int] = None
    max_len: Optional[int] = None
    alphabet: Optional[CaptchaAlphabet] = None
    language: Optional[WorkerLanguage] = None
    comment: Optional[str] = None

    def __post_init__(self):
        self._image_bytes = None
        self.get_image_bytes()

    def get_image_bytes(self) -> bytes:
        """ Bytes image """

        if self._image_bytes is None:
            if isinstance(self.image, bytes):
                self._image_bytes = self.image  # type: ignore
            elif isinstance(self.image, (io.RawIOBase, io.BufferedIOBase)):
                self._image_bytes = self.image.read()
            elif isinstance(self.image, pathlib.Path):
                self._image_bytes = self.image.read_bytes()

            # check image type
            self.get_image_type()

        return self._image_bytes

    def get_image_base64(self) -> bytes:
        """ BASE64 image """

        return base64.b64encode(self.get_image_bytes())

    def get_image_type(self) -> str:
        """ Get type of image file/data """

        image_type = imghdr.what(None, h=self._image_bytes)

        if not image_type:
            raise BadInputDataError("Unable to recognize image type!")
        return image_type


@enforce_types
@dataclass
class ImageCaptchaSolution(BaseCaptchaSolution):
    """ Image CAPTCHA solution """

    text: str



# -*- coding: UTF-8 -*-
"""
rucaptcha.com service
"""

# pylint: disable=unused-import
from .twocaptcha import (
    Service as Service2Captcha, GetBalanceRequest, GetStatusRequest,
    ReportGoodRequest, ReportBadRequest,
    ImageCaptchaTaskRequest, ImageCaptchaSolutionRequest,
    RecaptchaV2TaskRequest, RecaptchaV2SolutionRequest,
    RecaptchaV3TaskRequest, RecaptchaV3SolutionRequest,
    TextCaptchaTaskRequest, TextCaptchaSolutionRequest,
    FunCaptchaTaskRequest, FunCaptchaSolutionRequest,
    KeyCaptchaTaskRequest, KeyCaptchaSolutionRequest,
    GeeTestTaskRequest, GeeTestSolutionRequest,
    GeeTestV4TaskRequest, GeeTestV4SolutionRequest,
    HCaptchaTaskRequest, HCaptchaSolutionRequest,
    CapyPuzzleTaskRequest, CapyPuzzleSolutionRequest,
    TikTokCaptchaTaskRequest, TikTokCaptchaSolutionRequest
)

__all__ = [
    'Service', 'GetBalanceRequest', 'GetStatusRequest',
    'ReportGoodRequest', 'ReportBadRequest',
    'ImageCaptchaTaskRequest', 'ImageCaptchaSolutionRequest',
    'RecaptchaV2TaskRequest', 'RecaptchaV2SolutionRequest',
    'RecaptchaV3TaskRequest', 'RecaptchaV3SolutionRequest',
    'TextCaptchaTaskRequest', 'TextCaptchaSolutionRequest',
    'FunCaptchaTaskRequest', 'FunCaptchaSolutionRequest',
    'KeyCaptchaTaskRequest', 'KeyCaptchaSolutionRequest',
    'GeeTestTaskRequest', 'GeeTestSolutionRequest',
    'GeeTestV4TaskRequest', 'GeeTestV4SolutionRequest',
    'HCaptchaTaskRequest', 'HCaptchaSolutionRequest',
    'CapyPuzzleTaskRequest', 'CapyPuzzleSolutionRequest',
    'TikTokCaptchaTaskRequest', 'TikTokCaptchaSolutionRequest'
]


class Service(Service2Captcha):
    """ Main service class for rucaptcha """

    BASE_URL = 'https://rucaptcha.com'



# -*- coding: UTF-8 -*-

from unicaps._captcha import CaptchaType
from unicaps._captcha.base import BaseCaptcha, BaseCaptchaSolution


def test_captcha_class_base(captcha_class):
    """
    Checks base class of Captcha class.
    """
    assert issubclass(captcha_class, BaseCaptcha)


def test_captcha_class_get_type(captcha_class):
    """
    Checks get_type() function of Captcha class.
    """
    assert isinstance(captcha_class.get_type(), CaptchaType)


def test_captcha_class_get_solution_class(captcha_class):
    """
    Checks get_solution_class() function of Captcha class.
    """
    assert issubclass(captcha_class.get_solution_class(), BaseCaptchaSolution)


def test_captcha_class_get_optional_data(captcha_instance):
    """
    Checks get_optional_data() function of Captcha class.
    """
    assert isinstance(captcha_instance.get_optional_data(), dict)


def test_captchasolution_class_get_type(captcha_class):
    """
    Checks get_type() function of CaptchaSolution class.
    """
    assert captcha_class.get_type() is captcha_class.get_solution_class().get_type()


def test_captchasolution_class_get_captcha_class(captcha_class):
    """
    Checks get_type() function of CaptchaSolution class.
    """
    assert captcha_class is captcha_class.get_solution_class().get_captcha_class()


def test_captchasolution_obj_get_string(captcha_class):
    """
    Checks get_type() function of CaptchaSolution class.
    """
    solution_class = captcha_class.get_solution_class()

    dc_fields = solution_class.__dataclass_fields__
    field_values = []
    for i, field in enumerate(dc_fields.values(), start=1):
        if field.type == str:
            value = f'test{i}'
        elif field.type == dict:
            value = {f'test{i}': f'test{i}'}
        else:
            value = None

        field_values.append(value)

    solution_obj = solution_class(*field_values)

    assert str(solution_obj) == '\n'.join(str(v) for v in field_values)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="unicaps",
    version="1.2.1",
    author="Sergey Scat",
    author_email="py.unicaps@gmail.com",
    description="Universal CAPTCHA Solver for humans",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sergey-scat/unicaps",
    packages=setuptools.find_packages(),
    install_requires=["httpx>=0.22.0", "enforce-typing>=1.0.0"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10"
    ],
    python_requires='>=3.7',
)



# -*- coding: UTF-8 -*-
"""
Service module tests
"""

import importlib
import inspect
from copy import deepcopy
from unittest import mock

import pytest

from unicaps._captcha import CaptchaType
from unicaps._service import SOLVING_SERVICE

from data.data import (BASE_TASK_REQUEST_DATA, INPUT_TEST_DATA_FOR_TASK_PREPARE_FUNC,
                       OUTPUT_TEST_DATA_FOR_TASK_PREPARE_FUNC,
                       INPUT_TEST_LIST_FOR_TASK_PARSE_RESPONSE_FUNC,
                       INPUT_TEST_DATA_FOR_TASK_PARSE_RESPONSE_FUNC,
                       OUTPUT_TEST_DATA_FOR_TASK_PARSE_RESPONSE_FUNC,
                       INPUT_TEST_DATA_FOR_TASK_PARSE_RESPONSE_FUNC_WITH_EXC,
                       OUTPUT_TEST_DATA_FOR_TASK_PARSE_RESPONSE_FUNC_WITH_EXC)
from _helpers import dict_update

# dict of captcha services with lists of supported captchas
SERVICE_MODULES_FOR_TEST = {
    'anti_captcha': (CaptchaType.IMAGE, CaptchaType.RECAPTCHAV2, CaptchaType.RECAPTCHAV3,
                     CaptchaType.FUNCAPTCHA, CaptchaType.GEETEST, CaptchaType.GEETESTV4,
                     CaptchaType.HCAPTCHA),
    'twocaptcha': (CaptchaType.IMAGE, CaptchaType.TEXT, CaptchaType.RECAPTCHAV2,
                   CaptchaType.RECAPTCHAV3, CaptchaType.FUNCAPTCHA, CaptchaType.KEYCAPTCHA,
                   CaptchaType.GEETEST, CaptchaType.GEETESTV4, CaptchaType.HCAPTCHA,
                   CaptchaType.CAPY, CaptchaType.TIKTOK),
    'rucaptcha': (CaptchaType.IMAGE, CaptchaType.TEXT, CaptchaType.RECAPTCHAV2,
                  CaptchaType.RECAPTCHAV3, CaptchaType.FUNCAPTCHA, CaptchaType.KEYCAPTCHA,
                  CaptchaType.GEETEST, CaptchaType.GEETESTV4, CaptchaType.HCAPTCHA,
                  CaptchaType.CAPY, CaptchaType.TIKTOK),
    'azcaptcha': (CaptchaType.IMAGE, CaptchaType.RECAPTCHAV2, CaptchaType.RECAPTCHAV3,
                  CaptchaType.HCAPTCHA, CaptchaType.FUNCAPTCHA),
    'cptch_net': (CaptchaType.IMAGE, CaptchaType.RECAPTCHAV2, CaptchaType.RECAPTCHAV3),
    'deathbycaptcha': (CaptchaType.IMAGE, CaptchaType.RECAPTCHAV2, CaptchaType.RECAPTCHAV3,
                       CaptchaType.HCAPTCHA, CaptchaType.FUNCAPTCHA)
}
BASE_REQUESTS = ('GetBalance', 'GetStatus', 'ReportGood', 'ReportBad')
TASK_REQUEST_PREPARE_PARAMS = ('self', 'captcha', 'proxy', 'user_agent', 'cookies')
SOLUTION_REQUEST_PREPARE_PARAMS = ('self', 'task')
REQUEST_PARSE_RESPONSE_PARAMS = ('self', 'response')


@pytest.fixture(scope="module", params=SERVICE_MODULES_FOR_TEST)
def service_module(request):
    return importlib.import_module('unicaps._service.' + request.param)


@pytest.fixture(scope="module")
def service_instance(service_module):
    return getattr(service_module, "Service")("test")


@pytest.fixture(scope="module")
def json_response_obj():
    def get_obj(ret_value):
        obj = mock.Mock()
        obj.json = lambda: ret_value
        return obj
    return get_obj


def check_if_class_is_present(module, class_name, is_not=False):
    if is_not:
        assert not hasattr(module, class_name), \
               f'{class_name} is found in service module {module.__name__}'
    else:
        assert hasattr(module, class_name), \
               f'{class_name} is not found in service module {module.__name__}'


def is_captcha_supported(service_module, captcha_type):
    module_name = service_module.__name__.split('.')[-1]
    if captcha_type in SERVICE_MODULES_FOR_TEST[module_name]:
        return True
    return False


def get_request(service_module, captcha_type, req_type='Request'):
    request_name = captcha_type.value + req_type
    return request_name, getattr(service_module, request_name)


def test_if_service_class_is_present(service_module):
    """ Checks if Service() class is declared in the module """

    check_if_class_is_present(service_module, 'Service')


@pytest.mark.parametrize("req", BASE_REQUESTS)
def test_if_base_request_is_present(service_module, req):
    """ Checks if all of the base requests are present in the module file """

    check_if_class_is_present(service_module, req + 'Request')


@pytest.mark.parametrize("captcha_type", CaptchaType)
def test_if_task_request_is_present(service_module, captcha_type):
    """ Checks if all of the task requests are present in the module file """

    check_if_class_is_present(
        service_module,
        captcha_type.value + 'TaskRequest',
        not is_captcha_supported(service_module, captcha_type)
    )


@pytest.mark.parametrize("captcha_type", CaptchaType)
def test_if_solution_request_is_present(service_module, captcha_type):
    """ Checks if all of the solution requests are present in the module file """

    check_if_class_is_present(
        service_module,
        captcha_type.value + 'SolutionRequest',
        not is_captcha_supported(service_module, captcha_type)
    )


@pytest.mark.parametrize("req", BASE_REQUESTS)
def test_base_request_signature_of_prepare_func(service_module, req):
    """ Checks signature of the <captcha>TaskRequest.prepare() function """

    request_name = req + 'Request'
    request_class = getattr(service_module, request_name)

    if req in ('ReportGood', 'ReportBad'):
        standard = ('self', 'solved_captcha')
    else:
        standard = ('self',)

    params = tuple(inspect.signature(request_class.prepare).parameters)
    assert params == standard, \
           f"Incorrect signature of {request_name}.prepare() func: {', '.join(params)}"


@pytest.mark.parametrize("captcha_type", CaptchaType)
def test_task_request_signature_of_prepare_func(service_module, captcha_type):
    """ Checks signature of the <captcha>TaskRequest.prepare() function """

    if is_captcha_supported(service_module, captcha_type):
        request_name, request_class = get_request(service_module, captcha_type, 'TaskRequest')

        params = tuple(inspect.signature(request_class.prepare).parameters)
        assert params == TASK_REQUEST_PREPARE_PARAMS, \
               f"Incorrect signature of {request_name}.prepare() func: {', '.join(params)}"


@pytest.mark.parametrize("captcha_type", CaptchaType)
def test_solution_request_signature_of_prepare_func(service_module, captcha_type):
    """ Checks signature of the <captcha>TaskRequest.prepare() function """

    if is_captcha_supported(service_module, captcha_type):
        request_name, request_class = get_request(service_module, captcha_type, 'SolutionRequest')

        params = tuple(inspect.signature(request_class.prepare).parameters)
        assert params == SOLUTION_REQUEST_PREPARE_PARAMS, \
               f"Incorrect signature of {request_name}.prepare() func: {', '.join(params)}"


@pytest.mark.parametrize("req", BASE_REQUESTS)
def test_base_request_signature_of_parse_response_func(service_module, req):
    """ Checks signature of the <captcha>TaskRequest.prepare() function """

    request_name = req + 'Request'
    request_class = getattr(service_module, request_name)

    params = tuple(inspect.signature(request_class.parse_response).parameters)
    assert params == REQUEST_PARSE_RESPONSE_PARAMS, \
           f"Incorrect signature of {request_name}.prepare() func: {', '.join(params)}"


@pytest.mark.parametrize("captcha_type, req_type",
                         [(t, r) for t in CaptchaType for r in ('TaskRequest', 'SolutionRequest')])
def test_captcha_request_signature_of_parse_response_func(service_module, captcha_type, req_type):
    """
    Checks signature of the <captcha>TaskRequest.parse_response() and
    <captcha>SolutionRequest.parse_response() functions.
    """

    if not is_captcha_supported(service_module, captcha_type):
        pytest.skip("CAPTCHA is not supported!")

    request_name, request_class = get_request(service_module, captcha_type, req_type)
    params = tuple(inspect.signature(request_class.parse_response).parameters)
    assert params == REQUEST_PARSE_RESPONSE_PARAMS, \
           f"Incorrect signature of {request_name}.parse_response() func: {', '.join(params)}"


@pytest.mark.parametrize("test_id,input_data", INPUT_TEST_DATA_FOR_TASK_PREPARE_FUNC.items())
def test_task_request_return_value_of_prepare_func(service_module, test_id, input_data):
    service_type = {v: k for k, v in SOLVING_SERVICE.items()}[service_module]
    captcha_type = input_data[0].get_type()

    if not is_captcha_supported(service_module, captcha_type):
        pytest.skip("CAPTCHA is not supported!")

    request_name, request_class = get_request(service_module, captcha_type, 'TaskRequest')

    # service instance
    service = service_module.Service('test')

    # request instance
    request_instance = request_class(service)
    result_dict = request_instance.prepare(*input_data)

    standard_result = deepcopy(BASE_TASK_REQUEST_DATA[service_type])
    dict_update(standard_result, OUTPUT_TEST_DATA_FOR_TASK_PREPARE_FUNC[service_type][test_id])

    assert result_dict == standard_result


@pytest.mark.parametrize("test_id,captcha_type",
                         INPUT_TEST_LIST_FOR_TASK_PARSE_RESPONSE_FUNC.items())
def test_task_request_return_value_of_parse_response_func(service_module, test_id, captcha_type):
    service_type = {v: k for k, v in SOLVING_SERVICE.items()}[service_module]

    if not is_captcha_supported(service_module, captcha_type):
        pytest.skip("CAPTCHA is not supported!")

    request_name, request_class = get_request(service_module, captcha_type, 'TaskRequest')

    # service instance
    service = service_module.Service('test')

    input_data = INPUT_TEST_DATA_FOR_TASK_PARSE_RESPONSE_FUNC[service_type][test_id]

    # request instance
    request_instance = request_class(service)
    result_dict = request_instance.parse_response(input_data)

    standard_result = deepcopy(OUTPUT_TEST_DATA_FOR_TASK_PARSE_RESPONSE_FUNC[service_type][test_id])
    assert result_dict == standard_result


@pytest.mark.parametrize("test_id,captcha_type,exc_type",
                         [(i, c, e) for i, e in
                          OUTPUT_TEST_DATA_FOR_TASK_PARSE_RESPONSE_FUNC_WITH_EXC.items()
                          for c in CaptchaType])
def test_task_request_exception_of_parse_response_func(service_module, test_id, captcha_type,
                                                       exc_type):
    service_type = {v: k for k, v in SOLVING_SERVICE.items()}[service_module]

    if not is_captcha_supported(service_module, captcha_type):
        pytest.skip("CAPTCHA is not supported!")

    request_name, request_class = get_request(service_module, captcha_type, 'TaskRequest')

    # service instance
    service = service_module.Service('test')

    # get input data, skip test
    input_data = INPUT_TEST_DATA_FOR_TASK_PARSE_RESPONSE_FUNC_WITH_EXC[service_type][test_id]
    if not input_data:
        pytest.skip("The service doesn't support the testing exception!")

    # request instance
    request_instance = request_class(service)

    expected_exception = (
        OUTPUT_TEST_DATA_FOR_TASK_PARSE_RESPONSE_FUNC_WITH_EXC[test_id]
    )
    with pytest.raises(expected_exception):
        request_instance.parse_response(input_data)



# -*- coding: UTF-8 -*-
"""
Certain services related stuff
"""

import enum
# pylint: disable=import-self
from . import (
    anti_captcha, azcaptcha, captcha_guru, cptch_net, deathbycaptcha, rucaptcha, twocaptcha
)


class CaptchaSolvingService(enum.Enum):
    """ CAPTCHA solving service enumeration """

    ANTI_CAPTCHA = "anti-captcha.com"
    AZCAPTCHA = "azcaptcha.com"
    CAPTCHA_GURU = "captcha.guru"
    CPTCH_NET = "cptch.net"
    DEATHBYCAPTCHA = "deathbycaptcha.com"
    RUCAPTCHA = "rucaptcha.com"
    TWOCAPTCHA = "2captcha.com"


# supported CAPTCHA solving services
SOLVING_SERVICE = {
    CaptchaSolvingService.ANTI_CAPTCHA: anti_captcha,
    CaptchaSolvingService.AZCAPTCHA: azcaptcha,
    CaptchaSolvingService.CAPTCHA_GURU: captcha_guru,
    CaptchaSolvingService.CPTCH_NET: cptch_net,
    CaptchaSolvingService.DEATHBYCAPTCHA: deathbycaptcha,
    CaptchaSolvingService.RUCAPTCHA: rucaptcha,
    CaptchaSolvingService.TWOCAPTCHA: twocaptcha
}



# -*- coding: UTF-8 -*-
"""
CaptchaSolver class
"""
import io
import pathlib
from typing import Union

from .captcha import (
    ImageCaptcha, TextCaptcha, RecaptchaV2, RecaptchaV3, HCaptcha, FunCaptcha, KeyCaptcha, GeeTest,
    GeeTestV4, CapyPuzzle, TikTokCaptcha
)
from ._captcha.base import BaseCaptcha  # type: ignore
from ._service import CaptchaSolvingService, SOLVING_SERVICE
from ._service.base import SolvedCaptcha, CaptchaTask


class CaptchaSolver:
    """Main captcha solver :class:`CaptchaSolver <CaptchaSolver>` object.

    :param service_name: captcha solving service to use (enum CaptchaSolvingService or str).
    :param api_key: API key to access the solving service.
    """

    def __init__(self, service_name: Union[CaptchaSolvingService, str], api_key: str):
        # check service_name
        if isinstance(service_name, CaptchaSolvingService):
            self.service_name = service_name
        elif isinstance(service_name, str):
            try:
                self.service_name = CaptchaSolvingService(service_name)
            except ValueError as exc:
                raise ValueError(
                    f"'{service_name}' is not a valid CaptchaSolvingService. "
                    "Please use one of the following values: " + ', '.join(
                        [f"'{i.value}'" for i in CaptchaSolvingService]
                    )
                ) from exc
        else:
            raise ValueError(
                '"service_name" param must be an instance of str or CaptchaSolvingService!'
            )

        self.api_key = api_key
        self._service = SOLVING_SERVICE[self.service_name].Service(api_key)  # type: ignore

    def _solve_captcha(self, captcha_class, *args, **kwargs):
        proxy = kwargs.pop('proxy') if 'proxy' in kwargs else None
        user_agent = kwargs.pop('user_agent') if 'user_agent' in kwargs else None
        cookies = kwargs.pop('cookies') if 'cookies' in kwargs else None

        return self._service.solve_captcha(
            captcha_class(*args, **kwargs),
            proxy=proxy,
            user_agent=user_agent,
            cookies=cookies
        )

    def solve_image_captcha(self,
                            image: Union[bytes, io.RawIOBase, io.BufferedIOBase, pathlib.Path],
                            **kwargs) -> SolvedCaptcha:
        r"""Solves image CAPTCHA.

        :param image: binary file, bytes or pathlib.Path object containing image with CAPTCHA
        :param char_type: (optional) Character type.
        :param is_phrase: (optional) Boolean. True if CAPTCHA contains more than one word.
        :param is_case_sensitive: (optional) Boolean.
        :param is_math: (optional) Boolean. True if CAPTCHA requires calculation.
        :param min_len: (optional) Integer. Minimum length of the CAPTCHA's text.
        :param max_len: (optional) Integer. Maximum length of the CAPTCHA's text.
        :param alphabet: (optional) Alphabet used in the CAPTCHA.
        :param language: (optional) Language.
        :param comment: (optional) String. Text instructions for worker.
        :return: :class:`SolvedCaptcha <SolvedCaptcha>` object
        :rtype: unicaps.SolvedCaptcha
        """

        return self._solve_captcha(ImageCaptcha, image, **kwargs)

    def solve_text_captcha(self, text: str, **kwargs) -> SolvedCaptcha:
        r"""Solves text CAPTCHA.

        :param text: String with text captcha task.
        :param alphabet: (optional) Alphabet used in the CAPTCHA.
        :param language: (optional) Language.
        :return: :class:`SolvedCaptcha <SolvedCaptcha>` object
        :rtype: unicaps.SolvedCaptcha
        """
        return self._solve_captcha(TextCaptcha, text, **kwargs)

    def solve_recaptcha_v2(self, site_key: str, page_url: str, **kwargs) -> SolvedCaptcha:
        r"""Solves reCAPTCHA v2.

        :param site_key: Value of "data-sitekey" (or "k") parameter.
        :param page_url: Full URL of the page with CAPTCHA.
        :param is_invisible: (optional) Invisible reCAPTCHA flag.
        :param is_enterprise: (optional) reCAPTCHA Enterprise flag.
        :param data_s: (optional) Value of "data-s" parameter.
        :param api_domain: (optional) Domain used to load the captcha: google.com or recaptcha.net.
        :param proxy: (optional) Proxy to use while solving the CAPTCHA.
        :param user_agent: (optional) User-Agent to use while solving the CAPTCHA.
        :param cookies: (optional) Cookies to use while solving the CAPTCHA.
        :return: :class:`SolvedCaptcha <SolvedCaptcha>` object
        :rtype: unicaps.SolvedCaptcha
        """
        return self._solve_captcha(RecaptchaV2, site_key, page_url, **kwargs)

    def solve_recaptcha_v3(self, site_key: str, page_url: str, **kwargs) -> SolvedCaptcha:
        r"""Solves reCAPTCHA v3.

        :param site_key: Value of "render" parameter.
        :param page_url: Full URL of the page with CAPTCHA.
        :param is_enterprise: (optional) reCAPTCHA Enterprise flag (default: False).
        :param action: (optional) Widget action value.
        :param min_score: (optional) Filters a worker with corresponding score.
        :param api_domain: (optional) Domain used to load the captcha: google.com or recaptcha.net.
        :param proxy: (optional) Proxy to use while solving the CAPTCHA.
        :param user_agent: (optional) User-Agent to use while solving the CAPTCHA.
        :param cookies: (optional) Cookies to use while solving the CAPTCHA.
        :return: :class:`SolvedCaptcha <SolvedCaptcha>` object
        :rtype: unicaps.SolvedCaptcha
        """
        return self._solve_captcha(RecaptchaV3, site_key, page_url, **kwargs)

    def solve_hcaptcha(self, site_key: str, page_url: str, **kwargs) -> SolvedCaptcha:
        r"""Solves hCaptcha.

        :param site_key: hCaptcha website key.
        :param page_url: Full URL of the page with CAPTCHA.
        :param is_invisible: (optional) Invisible hCaptcha flag (default: False).
        :param api_domain: (optional) API domain: hcaptcha.com or js.hcaptcha.com.
        :param proxy: (optional) Proxy to use while solving the CAPTCHA.
        :param user_agent: (optional) User-Agent to use while solving the CAPTCHA.
        :param cookies: (optional) Cookies to use while solving the CAPTCHA.
        :return: :class:`SolvedCaptcha <SolvedCaptcha>` object
        :rtype: unicaps.SolvedCaptcha
        """
        return self._solve_captcha(HCaptcha, site_key, page_url, **kwargs)

    def solve_funcaptcha(self, public_key: str, page_url: str, **kwargs) -> SolvedCaptcha:
        r"""Solves FunCaptcha.

        :param public_key: FunCaptcha public key.
        :param page_url: Full URL of the page with CAPTCHA.
        :param service_url: (optional) Service URL.
        :param no_js: (optional) Disable JavaScript.
        :param blob: (optional) The "blob" value of CAPTCHA.
        :param proxy: (optional) Proxy to use while solving the CAPTCHA.
        :param user_agent: (optional) User-Agent to use while solving the CAPTCHA.
        :param cookies: (optional) Cookies to use while solving the CAPTCHA.
        :return: :class:`SolvedCaptcha <SolvedCaptcha>` object
        :rtype: unicaps.SolvedCaptcha
        """
        return self._solve_captcha(FunCaptcha, public_key, page_url, **kwargs)

    def solve_keycaptcha(self, page_url: str, user_id: str, session_id: str, ws_sign: str,
                         ws_sign2: str, **kwargs) -> SolvedCaptcha:
        r"""Solves KeyCaptcha.

        :param page_url: Full URL of the page with CAPTCHA.
        :param user_id: Value of "s_s_c_user_id" parameter.
        :param session_id: Value of "s_s_c_session_id" parameter.
        :param ws_sign: Value of "s_s_c_web_server_sign" parameter.
        :param ws_sign2: Value of "s_s_c_web_server_sign2" parameter.
        :return: :class:`SolvedCaptcha <SolvedCaptcha>` object
        :rtype: unicaps.SolvedCaptcha
        """
        return self._solve_captcha(
            KeyCaptcha, page_url, user_id, session_id, ws_sign, ws_sign2, **kwargs
        )

    def solve_geetest(self, page_url: str, gt_key: str, challenge: str,
                      **kwargs) -> SolvedCaptcha:
        r"""Solves GeeTest.

        :param page_url: Full URL of the page with CAPTCHA.
        :param gt_key: Public website key (static).
        :param challenge: Dynamic challenge key.
        :param api_server: (optional) API domain
        :return: :class:`SolvedCaptcha <SolvedCaptcha>` object
        :rtype: unicaps.SolvedCaptcha
        """
        return self._solve_captcha(GeeTest, page_url, gt_key, challenge, **kwargs)

    def solve_geetest_v4(self, page_url: str, captcha_id: str, **kwargs) -> SolvedCaptcha:
        r"""Solves GeeTestV4.

        :param page_url: Full URL of the page with CAPTCHA.
        :param captcha_id: Value of captcha_id parameter you found on target website.
        :return: :class:`SolvedCaptcha <SolvedCaptcha>` object
        :rtype: unicaps.SolvedCaptcha
        """
        return self._solve_captcha(GeeTestV4, page_url, captcha_id, **kwargs)

    def solve_capy_puzzle(self, site_key: str, page_url: str, **kwargs) -> SolvedCaptcha:
        r"""Solves Capy Puzzle CAPTCHA.

        :param site_key: Public website key (static).
        :param page_url: Full URL of the page with CAPTCHA.
        :param api_server: (optional) The domain part of script URL you found on page.
        :param proxy: (optional) Proxy to use while solving the CAPTCHA.
        :param user_agent: (optional) User-Agent to use while solving the CAPTCHA.
        :param cookies: (optional) Cookies to use while solving the CAPTCHA.
        :return: :class:`SolvedCaptcha <SolvedCaptcha>` object
        :rtype: unicaps.SolvedCaptcha
        """
        return self._solve_captcha(CapyPuzzle, site_key, page_url, **kwargs)

    def solve_tiktok(self, page_url: str, **kwargs) -> SolvedCaptcha:
        r"""Solves TikTokCaptcha.

        :param page_url: Full URL of the page with CAPTCHA.
        :param aid: (optional) The aid parameter value for the page.
        :param host: (optional) The host parameter value for the page.
        :param proxy: (optional) Proxy to use while solving the CAPTCHA.
        :param user_agent: (optional) User-Agent to use while solving the CAPTCHA.
        :param cookies: (optional) Cookies to use while solving the CAPTCHA.
        :return: :class:`SolvedCaptcha <SolvedCaptcha>` object
        :rtype: unicaps.SolvedCaptcha
        """
        return self._solve_captcha(TikTokCaptcha, page_url, **kwargs)

    def create_task(self, captcha: BaseCaptcha) -> CaptchaTask:
        """Create task to solve CAPTCHA

        :param captcha: Captcha to solve.
        :return: :class:`CaptchaTask <CaptchaTask>` object
        :rtype: unicaps.CaptchaTask
        """
        return self._service.create_task(captcha)

    def get_balance(self) -> float:
        """Get account balance

        :return: :float:Balance amount
        :rtype: float
        """
        return self._service.get_balance()

    def get_status(self) -> bool:
        """Get service status

        :return: :bool:Service status
        :rtype: bool
        """
        return self._service.get_status()

    def close(self) -> None:
        """Close all connections"""
        self._service.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()



""" Version info """

__title__ = 'unicaps'
__description__ = 'Python CAPTCHA solving for Humans.'
__url__ = 'https://github.com/sergey-scat/unicaps'
__version__ = '1.2.1'
__author__ = 'Sergey Scat'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2020-2023 Sergey Scat'



# -*- coding: UTF-8 -*-
"""
TikTokCaptcha
"""

from dataclasses import dataclass
from typing import Optional

from enforce_typing import enforce_types  # type: ignore

from .base import BaseCaptcha, BaseCaptchaSolution


@enforce_types
@dataclass
class TikTokCaptcha(BaseCaptcha):
    """ TikTokCaptcha """

    page_url: str
    aid: Optional[int] = None
    host: Optional[str] = None


@enforce_types
@dataclass
class TikTokCaptchaSolution(BaseCaptchaSolution):
    """ TikTokCaptcha solution """

    cookies: dict



# -*- coding: UTF-8 -*-
"""
Base transport stuff
"""

from abc import ABC, abstractmethod
from typing import Optional, Any


class BaseRequest(ABC):
    """ Base request class """

    def __init__(self, service):
        # solving service instance
        self._service = service
        # source request data (not None if a request in process)
        self.source_data = None

    @abstractmethod
    def prepare(self, **kwargs) -> dict:
        """ Prepare request data """
        self.source_data = kwargs
        return {}

    @abstractmethod
    def parse_response(self, response: Any) -> dict:
        """ Parse response """
        if self.source_data is None:
            raise RuntimeError('The Request.prepare() method must be called first!')
        return {}

    def process_response(self, response: Any) -> dict:
        """ Parse response and clean source request data """
        response = self.parse_response(response)
        self.source_data = None
        return response


class BaseTransport(ABC):  # pylint: disable=too-few-public-methods
    """ Base transport class """

    def __init__(self, settings: Optional[dict] = None):
        self.settings = settings or {}

    @abstractmethod
    def _make_request(self, request_data: dict) -> Any:
        """ Abstract method to make a request """

    @abstractmethod
    async def _make_request_async(self, request_data: dict) -> Any:
        """ Abstract method to make a request """

    def make_request(self, request: BaseRequest, *args) -> dict:
        """ Makes a request to the service """
        response = self._make_request(request.prepare(*args))
        return request.process_response(response)

    async def make_request_async(self, request: BaseRequest, *args) -> dict:
        """ Makes a request to the service """
        response = await self._make_request_async(request.prepare(*args))
        return request.process_response(response)

    @abstractmethod
    def close(self):
        """ Close connections """

    @abstractmethod
    async def close_async(self):
        """ Close connections (async) """
